"""Tests for filtered span classification module
"""
# Standard
from typing import Iterable, List
import os
import tempfile

# Third Party
from pytest import approx
import pytest

# First Party
from caikit.core import data_model
from caikit.core.modules import ModuleBase, module
from caikit.interfaces.nlp.data_model import (
    TokenClassificationResult,
    TokenClassificationResults,
)
from caikit.interfaces.nlp.tasks import TokenClassificationTask

# Local
from caikit_nlp.modules.text_classification import SequenceClassification
from caikit_nlp.modules.token_classification import FilteredSpanClassification
from caikit_nlp.modules.tokenization.regex_sentence_splitter import (
    RegexSentenceSplitter,
)
from tests.fixtures import SEQ_CLASS_MODEL

## Setup ########################################################################

# Bootstrapped sequence classification model
BOOTSTRAPPED_SEQ_CLASS_MODEL = SequenceClassification.bootstrap(SEQ_CLASS_MODEL)
# Regex sentence splitter model
SENTENCE_TOKENIZER = RegexSentenceSplitter.bootstrap(
    "[^.!?\s][^.!?\n]*(?:[.!?](?!['\"]?\s|$)[^.!?]*)*[.!?]?['\"]?(?=\s|$)"
)

DOCUMENT = (
    "The quick brown fox jumps over the lazy dog. Once upon a time in a land far away."
)

# Token classifications in document
FOX_CLASS = TokenClassificationResult(
    start=16, end=19, word="fox", entity="animal", score=0.8
)
DOG_CLASS = TokenClassificationResult(
    start=40, end=43, word="dog", entity="animal", score=0.3
)
LAND_CLASS = TokenClassificationResult(
    start=22, end=26, word="land", entity="thing", score=0.7
)
TOK_CLASSIFICATION_RESULT = TokenClassificationResults(results=[FOX_CLASS, DOG_CLASS])

# NOTE: First test will test this separately
BOOTSTRAPPED_MODEL = FilteredSpanClassification.bootstrap(
    lang="en",
    tokenizer=SENTENCE_TOKENIZER,
    classifier=BOOTSTRAPPED_SEQ_CLASS_MODEL,
    default_threshold=0.5,
)

# Modules that already returns token classification for tests
@module(
    "44d61711-c64b-4774-a39f-a9f40f1fcff0",
    "FakeTokenClassificationModule",
    "0.0.1",
    task=TokenClassificationTask,
)
class FakeTokenClassificationModule(ModuleBase):
    # This returns results for the whole document
    def run(self, text: str) -> TokenClassificationResults:
        return TOK_CLASSIFICATION_RESULT

    def run_batch(self, texts: List[str]) -> List[TokenClassificationResults]:
        return [
            TOK_CLASSIFICATION_RESULT,
            TokenClassificationResults(results=[LAND_CLASS]),
        ]


class StreamFakeTokenClassificationModule(FakeTokenClassificationModule):
    # Make module return results per sentence
    def run(self, text: str) -> TokenClassificationResults:
        if "land" in text:
            return TokenClassificationResults(results=[LAND_CLASS])
        else:
            return TOK_CLASSIFICATION_RESULT


class EmptyResFakeTokenClassificationModule(FakeTokenClassificationModule):
    def run(self, text: str) -> TokenClassificationResults:
        return TokenClassificationResults(results=[])

    def run_batch(self, texts: List[str]) -> List[TokenClassificationResults]:
        return [
            TokenClassificationResults(results=[]),
            TokenClassificationResults(results=[]),
        ]


TOKEN_CLASSIFICATION_MODULE = FakeTokenClassificationModule()
STREAM_TOKEN_CLASSIFICATION_MODULE = StreamFakeTokenClassificationModule()
EMPTY_RES_TOKEN_CLASSIFICATION_MODULE = EmptyResFakeTokenClassificationModule()

## Tests ########################################################################


def test_bootstrap_run():
    """Check if we can bootstrap and run span classification models with min arguments"""
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=BOOTSTRAPPED_SEQ_CLASS_MODEL,
        default_threshold=0.5,
    )
    token_classification_result = model.run(DOCUMENT)
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert len(token_classification_result.results) == 2  # 2 results over 0.5 expected
    assert isinstance(token_classification_result.results[0], TokenClassificationResult)
    first_result = token_classification_result.results[0]
    assert first_result.start == 0
    assert first_result.end == 44
    assert first_result.word == "The quick brown fox jumps over the lazy dog."
    assert first_result.entity == "LABEL_1"
    assert approx(first_result.score) == 0.50473803
    assert token_classification_result.results[1].entity == "LABEL_1"


def test_bootstrap_run_with_threshold():
    """Check if we can bootstrap span classification models with overriden threshold"""
    token_classification_result = BOOTSTRAPPED_MODEL.run(DOCUMENT, threshold=0.0)
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert (
        len(token_classification_result.results) == 4
    )  # 4 (all) results over 0.0 expected


def test_bootstrap_run_with_int_threshold():
    """Check if we can bootstrap span classification models with overriden int threshold"""
    token_classification_result = BOOTSTRAPPED_MODEL.run(DOCUMENT, threshold=0)
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert (
        len(token_classification_result.results) == 4
    )  # 4 (all) results over 0 expected


def test_bootstrap_run_with_optional_labels_to_output():
    """Check if we can run span classification models with labels_to_output"""
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=BOOTSTRAPPED_SEQ_CLASS_MODEL,
        default_threshold=0.5,
        labels_to_output=["LABEL_0"],
    )
    token_classification_result = model.run(DOCUMENT, threshold=0.0)
    # All results would be above threshold 0.0 but only return those corresponding to label
    assert len(token_classification_result.results) == 2
    first_result = token_classification_result.results[0]
    assert first_result.start == 0
    assert first_result.end == 44
    assert first_result.word == "The quick brown fox jumps over the lazy dog."
    assert first_result.entity == "LABEL_0"
    assert approx(first_result.score) == 0.49526197


def test_bootstrap_run_with_token_classification():
    """Check if we can run span classification models with classifier that does token classification"""
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=TOKEN_CLASSIFICATION_MODULE,
        default_threshold=0.5,
    )
    token_classification_result = model.run(DOCUMENT)
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert len(token_classification_result.results) == 2  # 2 results over 0.5 expected
    assert isinstance(token_classification_result.results[0], TokenClassificationResult)
    first_result = token_classification_result.results[0]
    assert first_result.start == 16
    assert first_result.end == 19
    assert first_result.word == "fox"
    assert first_result.entity == "animal"
    assert first_result.score == 0.8


def test_bootstrap_run_with_token_classification_no_results():
    """Check if we can run span classification models with classifier that does token classification
    but returns no results"""
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=EMPTY_RES_TOKEN_CLASSIFICATION_MODULE,
        default_threshold=0.5,
    )
    token_classification_result = model.run(DOCUMENT)
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert len(token_classification_result.results) == 0


def test_bootstrap_run_empty():
    """Check if span classification model can run with empty string"""
    token_classification_result = BOOTSTRAPPED_MODEL.run("")
    assert isinstance(token_classification_result, TokenClassificationResults)
    assert len(token_classification_result.results) == 0


def test_save_load_and_run_model():
    """Check if we can run a saved model successfully"""
    with tempfile.TemporaryDirectory() as model_dir:
        BOOTSTRAPPED_MODEL.save(model_dir)
        assert os.path.exists(os.path.join(model_dir, "config.yml"))
        assert os.path.exists(os.path.join(model_dir, "tokenizer"))
        assert os.path.exists(os.path.join(model_dir, "classification"))

        new_model = FilteredSpanClassification.load(model_dir)
        token_classification_result = new_model.run(DOCUMENT)
        assert isinstance(token_classification_result, TokenClassificationResults)
        assert (
            len(token_classification_result.results) == 2
        )  # 2 results over 0.5 expected


### Streaming tests ##############################################################


def test_run_bidi_stream_model():
    """Check if model prediction works as expected for bi-directional stream"""

    stream_input = data_model.DataStream.from_iterable(DOCUMENT)
    streaming_token_classification_result = BOOTSTRAPPED_MODEL.run_bidi_stream(
        stream_input
    )
    assert isinstance(streaming_token_classification_result, Iterable)
    # Convert to list to more easily check outputs
    result_list = list(streaming_token_classification_result)

    first_result = result_list[0].results[0]
    assert isinstance(first_result, TokenClassificationResult)
    assert first_result.start == 0
    assert first_result.end == 44
    assert first_result.word == "The quick brown fox jumps over the lazy dog."
    assert first_result.entity == "LABEL_1"
    assert approx(first_result.score) == 0.50473803

    # Check processed indices
    assert result_list[0].processed_index == 44
    assert result_list[1].processed_index == len(stream_input)

    # Assert total number of results should be equal to expected number of sentences
    expected_number_of_sentences = 2  # Sentence tokenizer returns 2 results
    count = len(result_list)
    assert count == expected_number_of_sentences


def test_run_bidi_stream_with_token_classification():
    """Check if model prediction with token classification
    works as expected for bi-directional stream"""

    stream_input = data_model.DataStream.from_iterable(DOCUMENT)
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=STREAM_TOKEN_CLASSIFICATION_MODULE,
        default_threshold=0.3,
    )
    streaming_token_classification_result = model.run_bidi_stream(stream_input)
    result_list = list(streaming_token_classification_result)
    # Convert to list to more easily check outputs
    first_result = result_list[0].results[0]
    assert isinstance(first_result, TokenClassificationResult)
    assert first_result.start == 16
    assert first_result.end == 19
    assert first_result.word == "fox"
    assert first_result.entity == "animal"
    assert first_result.score == 0.8

    # Check processed indices
    assert result_list[0].processed_index == 19  # token - fox
    assert result_list[1].processed_index == 43  # token - dog
    assert result_list[2].processed_index == 44  # end of first sentence
    assert result_list[3].processed_index == 71  # token - land
    assert result_list[4].processed_index == len(stream_input)  # end of second sentence

    # We expect 5 results here since there are 3 tokens found
    # and the rest of each of the 2 sentences
    # (to indicate the rest of the sentences are processed)
    expected_results = 5
    count = len(result_list)
    assert count == expected_results


def test_run_bidi_stream_with_token_classification_no_results():
    """Check if model prediction with token classification
    with no results works as expected for bi-directional stream"""
    stream_input = data_model.DataStream.from_iterable(DOCUMENT)
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=EMPTY_RES_TOKEN_CLASSIFICATION_MODULE,
        default_threshold=0.5,
    )
    streaming_token_classification_result = model.run_bidi_stream(stream_input)
    expected_results = 2  # Sentence tokenizer returns 2 results
    count = 0
    for result in streaming_token_classification_result:
        count += 1
        # Both sentences should not have results
        assert len(result.results) == 0

    # Assert total number of results should be equal to expected number of sentences
    assert count == expected_results


def test_run_bidi_stream_chunk_stream_input():
    """Check if model prediction with token classification
    with chunks of text input works as expected for bi-directional stream"""

    chunked_document_input = (
        "The quick brown fox jumps over the ",
        "lazy dog. Once upon a time in a land far away",
    )
    stream_input = data_model.DataStream.from_iterable(chunked_document_input)
    model = FilteredSpanClassification.bootstrap(
        lang="en",
        tokenizer=SENTENCE_TOKENIZER,
        classifier=STREAM_TOKEN_CLASSIFICATION_MODULE,
        default_threshold=0.3,
    )
    streaming_token_classification_result = model.run_bidi_stream(stream_input)
    result_list = list(streaming_token_classification_result)
    # Convert to list to more easily check outputs
    first_result = result_list[0].results[0]
    assert isinstance(first_result, TokenClassificationResult)
    assert first_result.start == 16
    assert first_result.end == 19
    assert first_result.word == "fox"
    assert first_result.entity == "animal"
    assert first_result.score == 0.8

    # Check processed indices
    assert result_list[0].processed_index == 19  # token - fox
    assert result_list[1].processed_index == 43  # token - dog
    assert result_list[2].processed_index == 44  # end of first sentence
    assert result_list[3].processed_index == 71  # token - land
    assert result_list[4].processed_index == 80  # end of second sentence

    # We expect 5 results here since there are 3 tokens found
    # and the rest of each of the 2 sentences
    # (to indicate the rest of the sentences are processed)
    expected_results = 5
    count = len(result_list)
    assert count == expected_results


def test_run_bidi_stream_with_multiple_spans_in_chunk():
    """Check if model prediction on stream with multiple sentences/spans
    works as expected for bi-directional stream"""
    doc_stream = (DOCUMENT, " I am another sentence.")
    stream_input = data_model.DataStream.from_iterable(doc_stream)

    streaming_token_classification_result = BOOTSTRAPPED_MODEL.run_bidi_stream(
        stream_input
    )
    assert isinstance(streaming_token_classification_result, Iterable)
    # Convert to list to more easily check outputs
    result_list = list(streaming_token_classification_result)

    first_result = result_list[0].results[0]
    assert isinstance(first_result, TokenClassificationResult)
    assert first_result.start == 0
    assert first_result.end == 44
    assert first_result.word == "The quick brown fox jumps over the lazy dog."
    assert first_result.entity == "LABEL_1"
    assert approx(first_result.score) == 0.50473803
    assert result_list[1].results[0].word == "Once upon a time in a land far away."
    assert result_list[2].results[0].word == "I am another sentence."

    # Check processed indices
    assert result_list[0].processed_index == 44  # end of first sentence (in DOC)
    assert result_list[1].processed_index == 81  # end of second sentence (in DOC)

    assert result_list[2].processed_index == 104  # end of third sentence (separate)

    # Assert total number of results should be equal to expected number of sentences
    expected_number_of_sentences = 3
    count = len(result_list)
    assert count == expected_number_of_sentences


def test_run_bidi_stream_empty():
    """Check if span classification model can run with empty string for streaming"""
    stream_input = data_model.DataStream.from_iterable("")
    streaming_token_classification_result = BOOTSTRAPPED_MODEL.run_bidi_stream(
        stream_input
    )
    assert isinstance(streaming_token_classification_result, Iterable)
    # Convert to list to more easily check outputs
    result_list = list(streaming_token_classification_result)
    assert len(result_list) == 1
    assert result_list[0].results == []
    assert result_list[0].processed_index == 0


def test_run_stream_vs_no_stream():
    """Check if model prediction on stream with multiple sentences/spans
    works as expected for bi-directional stream and gives expected span results
    as non-stream"""
    multiple_sentences = (
        "The dragon hoarded gold. The cow ate grass. What is happening? What a day!"
    )

    # Non-stream run
    nonstream_classification_result = BOOTSTRAPPED_MODEL.run(multiple_sentences)
    assert len(nonstream_classification_result.results) == 4
    assert nonstream_classification_result.results[0].word == "The dragon hoarded gold."
    assert nonstream_classification_result.results[0].start == 0
    assert nonstream_classification_result.results[0].end == 24
    assert nonstream_classification_result.results[3].word == "What a day!"
    assert nonstream_classification_result.results[3].start == 63
    assert nonstream_classification_result.results[3].end == 74

    # Char-based stream
    stream_input = data_model.DataStream.from_iterable(multiple_sentences)
    stream_classification_result = BOOTSTRAPPED_MODEL.run_bidi_stream(stream_input)
    # Convert to list to more easily check outputs
    result_list = list(stream_classification_result)
    assert len(result_list) == 4  # one per sentence
    assert result_list[0].processed_index == 24
    assert result_list[1].processed_index == 43
    assert result_list[2].processed_index == 62
    assert result_list[3].processed_index == 74

    # Chunk-based stream
    chunk_stream_input = data_model.DataStream.from_iterable((multiple_sentences,))
    chunk_stream_classification_result = BOOTSTRAPPED_MODEL.run_bidi_stream(
        chunk_stream_input
    )
    result_list = list(chunk_stream_classification_result)
    assert len(result_list) == 4  # one per sentence
    assert result_list[0].processed_index == 24
    assert result_list[1].processed_index == 43
    assert result_list[2].processed_index == 62
    assert result_list[3].processed_index == 74



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit_nlp.data_model import ExponentialDecayLengthPenalty

## Setup #########################################################################

dummy_exponential_decay_length_penalty = ExponentialDecayLengthPenalty(
    start_index=1, decay_factor=0.95
)

## Tests ########################################################################

### Exponential Decay Length Penalty
def test_exponential_decay_length_penalty_all_fields_accessible():
    assert dummy_exponential_decay_length_penalty.start_index == 1
    assert dummy_exponential_decay_length_penalty.decay_factor == 0.95


def test_sampling_parameters_from_proto_and_back():
    new = ExponentialDecayLengthPenalty.from_proto(
        dummy_exponential_decay_length_penalty.to_proto()
    )
    assert new.start_index == 1
    assert new.decay_factor == 0.95


def test_sampling_parameters_from_json_and_back():
    new = ExponentialDecayLengthPenalty.from_json(
        dummy_exponential_decay_length_penalty.to_json()
    )
    assert new.start_index == 1
    assert new.decay_factor == 0.95



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains utility classes for wrapping data streams as Torch datasets efficiently.
Caikit modules leveraging such wrappers can use them to internally leverage common approaches
and objects for training / evaluating PyTorch models built around DataStreams, e.g., PyTorch
DataLoaders, with minimal boilerplate.
"""

# Third Party
from torch.utils.data import IterableDataset

# First Party
from caikit.core.exceptions import error_handler
import alog

log = alog.use_channel("PEFT_PROMPT")
error = error_handler.get(log)


class SimpleIterableStreamWrapper(IterableDataset):
    """DataStream wrapper as an iterable PyTorch dataset; we use this to add
    compatability with PyTorch data loaders.
    """

    def __init__(self, stream, shuffle, buffer_size=None):
        error.type_check("<NLP12855513E>", bool, shuffle=shuffle)
        error.type_check(
            "<NLP12813713E>", int, buffer_size=buffer_size, allow_none=True
        )
        self.stream = stream
        self.shuffle = shuffle
        self.buffer_size = buffer_size
        # Load the whole data set in memory
        if self.shuffle and buffer_size is None:
            self.buffer_size = len(stream)
        log.debug("Shuffling enabled? {}".format(self.shuffle))
        log.debug("Shuffling buffer size: {}".format(self.buffer_size))

    def __iter__(self):

        # FIXME: We are currently not handling case where we have to work with
        # multiple workers, so currently duplicate data will get processed by
        # each worker.
        if self.shuffle:
            log.debug4("Reshuffling training data!")
            return iter(self.stream.shuffle(self.buffer_size))
        return iter(self.stream)
        # worker_info = get_worker_info()
        # if worker_info is None:  # single-process data loading, return the full iterator
        #     if self.shuffle:
        #         log.debug4("Reshuffling training data!")
        #         return iter(self.stream.shuffle(self.buffer_size))
        #     return iter(self.stream)

        # When num_workers > 0, each worker process will have a different copy of
        # the dataset object, so we configure each copy independently to avoid
        # having duplicate data returned from each worker
        # else:  # in a worker process
        # # split workload
        # per_worker = int(
        #     math.ceil((self.end - self.start) / float(worker_info.num_workers))
        # )
        # worker_id = worker_info.id
        # iter_start = self.start + worker_id * per_worker
        # iter_end = min(iter_start + per_worker, self.end)
        # return iter(range(iter_start, iter_end))

    def __len__(self):
        return len(self.stream)



"""This module holds references to common util functions and constants used for training
and evaluating models.
"""
# Standard
import os
import sys

# Hack for relative imports outside of containerized environments
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

# Standard
from collections import namedtuple
from shutil import which
from typing import Any, Callable, Tuple
import math
import random

# Third Party
import datasets
import evaluate
import numpy as np
import torch
import transformers

# First Party
from caikit_tgis_backend import TGISBackend
import alog
import caikit

# Local
from caikit_nlp.data_model import GenerationTrainRecord

# Silence noisy import time tensorflow warnings
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

ALOG_OPTS = {
    "filters": "datasets:off,urllib3:off,apscheduler:off,tzloc:off",
    "default_level": "error",
    "formatter": "pretty",
}

log = alog.use_channel("EXMPL_UTILS")


def configure_random_seed_and_logging():
    """Ensure that random experiments will be deterministic & set up default ALOG configuration."""
    seed = 1
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    transformers.set_seed(seed)
    # Default Alog configuration; this may be overridden in tests as needed,
    # but some of these libraries have super verbose logging
    alog.configure(**ALOG_OPTS)


def print_colored(print_obj: Any) -> None:
    """Print stuff important to our experiment in blue text.
    Args:
        print_obj: Any
            Object to be printed.
    """
    print("\033[94m{}\033[0m".format(print_obj))


### Common loading / model capabilities
kill_tgis_container_if_exists = lambda: os.system("./kill-text-generation-launcher.sh")


def get_distributed_model(model_path):
    """Get the distributed implementation of the same model."""
    has_text_gen = which("text-generation-launcher")
    if not which("text-generation-launcher"):
        print("Text generation server command not found; using Docker override")
        this_dir = os.path.dirname(os.path.abspath(__file__))
        os.environ["PATH"] += ":" + this_dir
        assert (
            which("text-generation-launcher") is not None
        ), "Text generation script not found!"
        # Kill the docker container for TGIS if it's currently running
        kill_tgis_container_if_exists()

    # TODO: Enforce validation that TGIS is mounting the same model type

    caikit.config.configure(
        config_dict={
            "model_management": {
                "initializers": {
                    "default": {
                        "config": {
                            "backend_priority": [{"type": TGISBackend.backend_type}]
                        }
                    }
                }
            }
        }
    )  # should not be necessary but just in case
    dist_model = caikit.load(model_path)
    # Sanity check; if we have an environment variable override for the model TGIS is using,
    # make sure that its suffix (base model name) aligns with what we have in our config.
    # NOTE: bloom-560m is the default here because that's the default model used in our
    # text generation server hack script.
    model_name_override = os.getenv("MODEL_NAME", "bigscience/bloom-560m")
    if hasattr(dist_model, "base_model_name"):
        loaded_base_model = dist_model.base_model_name
    else:
        loaded_base_model = dist_model.model_name
    if not model_name_override.endswith(loaded_base_model):
        log.error(
            "TGIS using model name: {} conflicts with base model name: {}; set env var MODEL_NAME to the correct base model!".format(
                model_name_override, loaded_base_model
            )
        )

    return dist_model


def load_model(is_distributed: bool, model_path: str):
    """Loads a model for evaluation, either as a local module, or in a containerized
    instance of TGIS.

    """
    # Ensure caikit_nlp is locally imported, otherwise it'll be missing in out registry
    # Local
    import caikit_nlp

    # Validate that this model is something we actually know how to load
    if is_distributed:
        return get_distributed_model(model_path)
    return caikit_nlp.load(model_path)


### Dataset specific loader funcs
def load_twitter_dataset(
    get_test_set_as_eval=False,
) -> Tuple[caikit.core.data_model.DataStream]:
    """Load the ought/raft twitter complaints dataset.

    Returns:
        Tuple(caikit.core.data_model.DataStream)
            DataStreams of GenerationTrainRecords to be leveraged for training, validation,
            and testing, respectively.
    """
    to_generation_fmt = lambda x: GenerationTrainRecord(
        input=x["Tweet text"], output=x["text_label"]
    )
    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in dataset[split]]
    )
    dataset_for_task = datasets.load_dataset("ought/raft", "twitter_complaints")
    # Note: The data preprocessing (train, val and test) is copied from the Apache 2.0 code under
    #       https://github.com/huggingface/peft/blob/main/examples/causal_language_modeling/peft_prompt_tuning_clm.ipynb
    classes = [k.replace("_", " ") for k in dataset_for_task["train"].features["Label"].names]  # type: ignore

    # Classes are ['Unlabeled', 'complaint', 'no complaint'] and given by integer => substitute int with NL label
    dataset = dataset_for_task.map(
        lambda x: {"text_label": [classes[label] for label in x["Label"]]},
        batched=True,
        num_proc=1,  # type: ignore
    )
    # This dataset has no validation data
    train_stream = build_stream("train")
    test_stream = build_stream("test")
    print_colored(
        "Warning: using train stream as validation; twitter data has no validation set!"
    )
    return (train_stream, train_stream, test_stream)


def load_cola_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the Glue Cola dataset."""
    to_generation_fmt = lambda x: GenerationTrainRecord(
        input=x["sentence"], output=str(x["label"])
    )
    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in dataset[split]]
    )
    dataset = datasets.load_dataset("glue", "cola")
    train_stream = build_stream("train")
    validation_stream = build_stream("validation")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def load_rte_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the Glue RTE dataset."""

    def to_generation_fmt(x):
        source_text = " ".join(
            ["sentence1:", x["sentence1"], "sentence2:", x["sentence2"]]
        )
        return GenerationTrainRecord(input=source_text, output=str(x["label"]))

    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in dataset[split]]
    )
    dataset = datasets.load_dataset("glue", "rte")
    train_stream = build_stream("train")
    validation_stream = build_stream("validation")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def load_mrpc_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the Glue MRPC dataset."""

    def to_generation_fmt(x):
        source_text = " ".join(
            ["sentence1:", x["sentence1"], "sentence2:", x["sentence2"]]
        )
        return GenerationTrainRecord(input=source_text, output=str(x["label"]))

    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in dataset[split]]
    )
    dataset = datasets.load_dataset("glue", "mrpc")
    train_stream = build_stream("train")
    validation_stream = build_stream("validation")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def load_financial_phrasebank_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the financial_phrasebank dataset."""

    def to_generation_fmt(x):
        return GenerationTrainRecord(input=x["sentence"], output=str(x["label"]))

    dataset = datasets.load_dataset("financial_phrasebank", "sentences_allagree")
    train_test_dataset = dataset["train"].train_test_split(test_size=0.1)
    # # Split the 10% test + valid into half test, half valid
    # test_valid = train_test_valid_dataset['test'].train_test_split(test=0.5)

    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in train_test_dataset[split]]
    )
    train_stream = build_stream("train")
    validation_stream = build_stream("test")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def load_billsum_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the billsum dataset."""

    def to_generation_fmt(x):
        return GenerationTrainRecord(input=x["text"], output=str(x["summary"]))

    dataset = datasets.load_dataset("billsum", split="ca_test")
    train_test_dataset = dataset.train_test_split(test_size=0.2)
    # # Split the 10% test + valid into half test, half valid
    # test_valid = train_test_valid_dataset['test'].train_test_split(test=0.5)

    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in train_test_dataset[split]]
    )
    train_stream = build_stream("train")
    validation_stream = build_stream("test")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def load_samsum_dataset() -> Tuple[caikit.core.data_model.DataStream]:
    """Load the samsum dataset."""

    def to_generation_fmt(x):
        return GenerationTrainRecord(input=x["dialogue"], output=str(x["summary"]))

    dataset = datasets.load_dataset("samsum")
    train_test_dataset = dataset.train_test_split(test_size=0.1)
    # # Split the 10% test + valid into half test, half valid
    # test_valid = train_test_valid_dataset['test'].train_test_split(test=0.5)

    build_stream = lambda split: caikit.core.data_model.DataStream.from_iterable(
        [to_generation_fmt(datum) for datum in train_test_dataset[split]]
    )
    train_stream = build_stream("train")
    validation_stream = build_stream("validation")
    test_stream = build_stream("test")
    return (train_stream, validation_stream, test_stream)


def get_wrapped_evaluate_metric(metric_name: str, convert_to_numeric: bool) -> Callable:
    """Wrapper for running metrics out of evaluate which operate on numeric arrays
    named predictions & references, respectively.

    Args:
        metric_name: str
            Name of the metric to be run.
        convert_to_numeric: bool
            Indicates whether kwargs needed to be converted to floats prior to evaluation.
    Returns:
        Callable
            Callable evaluate metric which takes lists of strings as inputs.
    """
    metric = evaluate.load(metric_name)
    if convert_to_numeric:
        return build_func_for_numeric_input_metric_func(metric.compute)
    return metric.compute


def build_func_for_numeric_input_metric_func(func) -> Callable:
    """Builds a function that coerces its kwargs to numerics prior to forwarding to
    an internal function for metric evaluation.
    Args:
        func: Callable
            Function to be wrapped.

    Returns:
        Callable
            Function we can call with numeric inputs.
    """

    def metric_func_with_str_args(predictions: Tuple[str], references: Tuple[str]):
        # Convert all model predictions & targets to numerics
        numeric_preds = [string_to_float(pred, strict=False) for pred in predictions]
        numeric_refs = [string_to_float(ref, strict=True) for ref in references]
        return func(predictions=numeric_preds, references=numeric_refs)

    return metric_func_with_str_args


def is_float(string: str) -> bool:
    """Determine if a string is castable into a float.

    Args:
        string: str
            Value we want to try to cast to a float.

    Returns:
        bool
            True if the input string is castable to a bool, False otherwise.
    """
    try:
        float(string)
        return True
    except ValueError:
        return False


def string_to_float(string: str, strict: bool, default: int = -1.0):
    """Converts string to float, using default when conversion not possible.
    Here, we explicitly enforce that provided strings are numerically castable.

    Args:
        string: str
            Input string to be cast to numeric.
        default: int
            Default numeric label idx to map garbage

    Returns:
        int
            Label index of the cast result, which is presumably somehow mapped to classification
            (for currently supported metrics).
    """
    if strict and not is_float(string):
        raise ValueError(
            "Unable to cast string: [{}] to float in strict mode".format(string)
        )
    # Otherwise fall back to direct casting, and just return the cast value for
    # correctly generated indices, or the default value for what is presumably garbage.
    try:
        return float(string)
    except ValueError:
        return default


# Global map of supported datasets; users are able to select one of these to train prompt
# vectors against. Each value in the map should contain a default verbalizer, a nonparametric
# loader func for grabbing a tuple of three (train, validation||None, test||None) datastreams,
# and an initialization text string to use for initialize prompt tuning if TEXT is selected
# as the initialization option.
DatasetInfo = namedtuple("DatasetInfo", ["verbalizer", "dataset_loader", "init_text"])
SUPPORTED_DATASETS = {
    "twitter_complaints": DatasetInfo(
        verbalizer="Tweet text : {{input}} Label : ",
        dataset_loader=load_twitter_dataset,
        init_text="Classify if the tweet is a complaint or not:",
    ),
    "glue/cola": DatasetInfo(
        verbalizer="cola { 0 : grammatically unacceptable, 1 : grammatically acceptable } sentence: {{input}}",
        dataset_loader=load_cola_dataset,
        init_text="Classify if the text is a grammatical English sentence or not:",
    ),
    "glue/rte": DatasetInfo(
        verbalizer="rte { 0 : entailment, 1 : not entailment } {{input}}",
        dataset_loader=load_rte_dataset,
        init_text="Recognize textual entailment:",
    ),
    "glue/mrpc": DatasetInfo(
        verbalizer="mrpc { 0 : not equivalent, 1 : equivalent } {{input}}",
        dataset_loader=load_mrpc_dataset,
        init_text="Determine if the sentences are semantically equivalent: ",
    ),
    "financial_phrasebank": DatasetInfo(
        verbalizer="{{input}}",
        dataset_loader=load_financial_phrasebank_dataset,
        init_text="Classify sentiment for each of the news articles: ",
    ),
    "billsum": DatasetInfo(
        verbalizer="{{input}}",
        dataset_loader=load_billsum_dataset,
        init_text="",
    ),
    "samsum": DatasetInfo(
        verbalizer="{{input}}",
        dataset_loader=load_samsum_dataset,
        init_text="",
    ),
}

# Supported metrics in huggingface's evaluate library.
MetricInfo = namedtuple("MetricInfo", ["metric_name", "convert_to_numeric"])
METRIC_INFOS = [
    MetricInfo(metric_name="accuracy", convert_to_numeric=True),
    MetricInfo(metric_name="matthews_correlation", convert_to_numeric=True),
    MetricInfo(metric_name="rouge", convert_to_numeric=False),
]
SUPPORTED_METRICS = {
    metric_info.metric_name: get_wrapped_evaluate_metric(
        metric_info.metric_name, metric_info.convert_to_numeric
    )
    for metric_info in METRIC_INFOS
}



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# First Party
from caikit.core.exceptions import error_handler
from caikit.interfaces.nlp.data_model import ClassificationTrainRecord
import alog

# Local
from ..data_model import GenerationTrainRecord

log = alog.use_channel("TASK_UTILS")
error = error_handler.get(log)


def convert_to_generation_record(train_record):
    if isinstance(train_record, GenerationTrainRecord):
        return train_record
    if isinstance(train_record, ClassificationTrainRecord):
        text = train_record.text
        labels = labels = ",".join(str(label) for label in train_record.labels)
        return GenerationTrainRecord(input=text, output=labels)
    error(
        "<NLP12517812E>",
        TypeError(
            "Unsupported instance type. \
            Only instances of datamodels ClassificationTrainRecord \
            and GenerationTrainRecord are supported"
        ),
    )



"""This script illustrates how to train a prompt tuned model using prompt tuning / MPT.
Supported tuning types:
- MultiPrompt Tuning
- Prompt Tuning

Supported model types:
- Causal LM
- Seq2Seq LM
"""
# Standard
from collections import namedtuple
from typing import Any, Tuple
import argparse
import os
import pathlib
import random
import shutil
import time

# Third Party
from peft.tuners.multitask_prompt_tuning import MultitaskPromptTuningInit
from peft.tuners.prompt_tuning import PromptTuningInit
from transformers import AutoConfig
from utils import (
    ALOG_OPTS,
    SUPPORTED_DATASETS,
    DatasetInfo,
    configure_random_seed_and_logging,
    print_colored,
)
import datasets

# First Party
from caikit.core.data_model import DataStream
import alog
import caikit

# Local
from caikit_nlp.data_model import GenerationTrainRecord, TuningConfig
from caikit_nlp.modules.text_generation.peft_prompt_tuning import (
    PeftPromptTuning,
    TuningType,
)
from caikit_nlp.resources.pretrained_model import (
    HFAutoCausalLM,
    HFAutoSeq2SeqLM,
    PretrainedModelBase,
)


def subsample_stream(
    train_stream: DataStream[GenerationTrainRecord], num_shots: int
) -> DataStream[GenerationTrainRecord]:
    """Given a training stream of length n, randomly extract num_shots <= n samples from it
    for use in few-shot learning.

    Args:
        train_stream: DataStream
            Full dataset to be used for training prior to few shot sampling.
        num_shots: int
            Number of samples to keep for use in training.

    Returns:
        DataStream[GenerationTrainRecord]
            Train subsampled stream of len(x) == num_shots
    """
    num_samples = len(train_stream)
    if num_shots > num_samples or num_shots <= 0:
        raise ValueError(
            "num_shots [{}] is less than 0 or exceeds train data size: [{}]".format(
                num_shots, num_samples
            )
        )
    # If we have the same number of samples as shots, just give the raw stream back
    elif num_shots == num_samples:
        return train_stream
    # Otherwise subsample the stream to condense its length; shuffle using
    # the whole stream as a buffer, and build a new stream from the result.
    # NOTE - this is not a great idea, but for now we do this, so that the sampling
    # is exactly the same as the original MPT code, since sampling the whole dataset
    # with a max buffer would load everything into memory anyway.
    shuffled_dataset = random.sample(list(train_stream), num_shots)
    return DataStream.from_iterable(shuffled_dataset)


def get_resource_type(model_name: str) -> PretrainedModelBase:
    """Given a model name, or a path to a model, get the resource type to be initialized.

    Args:
        model_name: str
            Model name or path to the model to be leveraged.

    Returns:
        type
            PretrainedModel subclass wrapping the loaded Transformer model, e.g.,
            a HFAutoCausalLM or HFAutoSeq2SeqLM. We return the type here so that
            we can initialize it later, after we show a nice experiment configuration.
    """
    try:
        model_type = AutoConfig.from_pretrained(model_name).model_type
    except OSError:
        raise ValueError("Failed to load model with name: {}".format(model_name))
    if model_type in HFAutoCausalLM.SUPPORTED_MODEL_TYPES:
        return HFAutoCausalLM
    elif model_type in HFAutoSeq2SeqLM.SUPPORTED_MODEL_TYPES:
        return HFAutoSeq2SeqLM
    raise NotImplementedError(
        "Provided is not supported for any supported resource type!"
    )


### Functions for arg parsing & validation
def parse_args() -> argparse.Namespace:
    """Parse command line arguments. Here, we set up each tuning task as a subparser
    to prevent the arguments from being too confusing. Common arguments, e.g., the number
    of virtual tokens, are added to all parsers.

    Returns:
        argparse.Namespace
            Parsed arguments to be leveraged for one prompt tuning application.
    """
    parser = argparse.ArgumentParser(
        description="Train prompt vectors on top of a text generation model.",
    )
    ### Args specific to subparsers, i.e., tuning / training arguments
    subparsers = parser.add_subparsers(
        help="The type of tuning to apply.", dest="tuning_type", required=True
    )
    # NOTE: These keys should line up with the TuningType enum values
    parser_multiprompt_tuning = subparsers.add_parser(
        "MULTITASK_PROMPT_TUNING",
        help="Train prompt vectors through Multitask Prompt Tuning.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_prompt_tuning = subparsers.add_parser(
        "PROMPT_TUNING",
        help="Train prompt vectors through Prompt Tuning.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = (
        parser_multiprompt_tuning,
        parser_prompt_tuning,
    )
    # Register all of the common args, as well as specific tuning args for subcommands
    register_common_arguments(subparsers)
    register_multitask_prompt_tuning_args(parser_multiprompt_tuning)
    register_prompt_tuning_args(parser_prompt_tuning)
    args = parser.parse_args()
    # Reconfigure logging level based on verbosity, while preserving filters etc.
    default_level = "debug" if args.verbose else "info"
    alog_settings = {**ALOG_OPTS, **{"default_level": default_level}}
    alog.configure(**alog_settings)
    # Validate common arg values
    validate_common_args(args)
    return args


def register_common_arguments(subparsers: Tuple[argparse.ArgumentParser]) -> None:
    """Registers common arguments intended to be shared across all subparsers.

    Args:
        subparsers: Tuple[argparse.ArgumentParser]
            Iterable of argument subparsers that should have common args.
    """
    for subparser in subparsers:
        subparser.add_argument(
            "--dataset",
            help="Dataset to use to train prompt vectors. Options: {}".format(
                list(SUPPORTED_DATASETS.keys())
            ),
            default="twitter_complaints",
        )
        subparser.add_argument(
            "--model_name",
            help="Name of base model or path to model to use to train prompt vectors",
            default="bigscience/bloom-560m",
        )
        subparser.add_argument(
            "--output_dir",
            help="Name of the directory that we want to export the model to",
            default="sample_prompt",
        )
        subparser.add_argument(
            "--prompt_only",
            help="Indicates that we do not need to export the full model, just the prompt vectors",
            action="store_true",
        )
        subparser.add_argument(
            "--verbose",
            help="If enabled, shows TQDM progress bars & debug logs",
            action="store_true",
        )
        subparser.add_argument(
            "--num_virtual_tokens",
            help="Number of virtual tokens to use per transformer submodule",
            type=int,
            default=8,
        )
        subparser.add_argument(
            "--num_epochs",
            help="Number of epochs to use for prompt tuning",
            type=int,
            default=10,
        )
        subparser.add_argument(
            "--learning_rate",
            help="Learning rate to use while training",
            type=float,
            default=3e-2,
        ),
        subparser.add_argument(
            "--num_shots",
            help="Number of training samples to use for few-shot learning",
            type=int,
            default=None,
        ),
        subparser.add_argument(
            "--batch_size", help="Batch size to use while training", type=int, default=8
        )
        subparser.add_argument(
            "--max_source_length",
            help="Maximum source sequence length.",
            default=256,
            type=int,
        )
        subparser.add_argument(
            "--max_target_length",
            help="Maximum target sequence length.",
            default=128,
            type=int,
        )
        subparser.add_argument(
            "--accumulate_steps",
            help="Gradient accumulation steps",
            default=1,
            type=int,
        )
        subparser.add_argument(
            "--torch_dtype",
            help="Torch dtype to be used for training",
            default="float32",
            choices=["float16", "bfloat16", "float32"],
        )


def register_multitask_prompt_tuning_args(subparser: argparse.ArgumentParser):
    """Register additional configuration options for MP(rompt)T subtask.

    Args:
        subparser: argparser.ArgumentParser
            Configuration options for multitask prompt tuning specifically.
    """
    subparser.add_argument(
        "--prompt_tuning_init",
        help="Initialization method to be used for multitask prompt tuning",
        choices=[x.value for x in MultitaskPromptTuningInit],
        default="TEXT",
    )
    subparser.add_argument(
        "--prompt_tuning_init_source_model",
        help="Path to source prompts to initialize multitask prompt tuning",
        type=pathlib.Path,
    )


def register_prompt_tuning_args(subparser: argparse.ArgumentParser):
    """Register additional configuration options for prompt tuning subtask.

    Args:
        subparser: argparser.ArgumentParser
            Configuration options for prompt tuning specifically.
    """
    subparser.add_argument(
        "--prompt_tuning_init",
        help="Initialization method to be used for prompt tuning",
        choices=[x.name for x in PromptTuningInit],
        default="TEXT",
    )


def validate_common_args(args: argparse.Namespace):
    """Validates common arguments to ensure values make sense; here, we only validate things that
    are not (or should not) be handled within the module.

    Args:
        args: argparse.Namespace
            Parsed args corresponding to one tuning task.
    """
    # Validate that the dataset is one of our allowed values
    if args.dataset not in SUPPORTED_DATASETS:
        raise KeyError(
            "[{}] is not a supported dataset; see --help for options.".format(
                args.dataset
            )
        )
    # Purge our output directory if one already exists.
    if os.path.isdir(args.output_dir):
        print("Existing model directory found; purging it now.")
        shutil.rmtree(args.output_dir)


def build_tuning_config(args: argparse.Namespace, dataset_info: DatasetInfo):
    """Builds the tuning config for this tuning task.

    Args:
        args: argparse.Namespace
            Parsed args corresponding to one tuning task.
        dataset_info: DatasetInfo
            Dataset information, including text to be used to initialize prompt tuning if
            that's the selected initialization scheme.

    Returns
        TuningConfig
            Tuning config object to be provided at .train() time.
    """
    # NOTE: the block itself already does filtering on our options, but since we separate things
    # through the CLI, we build the Tuning Config as accurately as possible so that the options
    # can be shown in the console.
    base_kwargs = {
        "num_virtual_tokens": args.num_virtual_tokens,
        "prompt_tuning_init_method": args.prompt_tuning_init,
    }
    # Add the initialization text only if we actually initialize with text
    if args.prompt_tuning_init == "TEXT":
        base_kwargs["prompt_tuning_init_text"] = dataset_info.init_text
    if (
        args.tuning_type == "MULTITASK_PROMPT_TUNING"
        and args.prompt_tuning_init_source_model
    ):
        if not args.prompt_tuning_init_source_model.exists():
            raise FileNotFoundError(
                "Provided prompt tuning init source does not exist!"
            )
        base_kwargs["prompt_tuning_init_source_model"] = str(
            args.prompt_tuning_init_source_model
        )
    return TuningConfig(**base_kwargs)


def show_experiment_configuration(args, dataset_info, model_type) -> None:
    """Show the complete configuration for this experiment, i.e., the model info,
    the resource type we built, the training params, metadata about the dataset where
    possible, and so on.

    Args:
        args: argparse.Namespace
            Parsed args corresponding to one tuning task.
        dataset_info: DatasetInfo
            Dataset information, including text to be used to initialize prompt tuning if
            that's the selected initialization scheme.
        model_type: type
            Resource class corresponding to the base model.
    """
    text_init_substr = (
        " |- Prompt Tuning initialization Text: [{}]".format(dataset_info.init_text)
        if args.prompt_tuning_init == "TEXT"
        else ""
    )
    print_strs = [
        "Experiment Configuration",
        "- Model Name: [{}]".format(args.model_name),
        " |- Inferred Model Resource Type: [{}]".format(model_type),
        "- Tuning Type: [{}]".format(args.tuning_type),
        "- Prompt Tuning Initialization Type [{}]".format(args.prompt_tuning_init),
        "- Number of Virtual Tokens: [{}]".format(args.num_virtual_tokens),
        text_init_substr,
        "- Dataset: [{}]".format(args.dataset),
        "- Verbalizer: [{}]".format(dataset_info.verbalizer),
        "- Number of Epochs: [{}]".format(args.num_epochs),
        "- Learning Rate: [{}]".format(args.learning_rate),
        "- Batch Size: [{}]".format(args.batch_size),
        "- Output Directory: [{}]".format(args.output_dir),
        "- Exporting prompt only: [{}]".format(args.prompt_only),
        "- Number of shots: [{}]".format(args.num_shots),
        "- Maximum source sequence length: [{}]".format(args.max_source_length),
        "- Maximum target sequence length: [{}]".format(args.max_target_length),
        "- Gradient accumulation steps: [{}]".format(args.accumulate_steps),
    ]
    # Log and sleep for a few seconds in case people actually want to read this...
    print_colored("\n".join([print_str for print_str in print_strs if print_str]))


if __name__ == "__main__":
    configure_random_seed_and_logging()
    args = parse_args()
    model_type = get_resource_type(args.model_name)
    # Unpack the dataset dictionary into a loaded dataset & verbalizer
    dataset_info = SUPPORTED_DATASETS[args.dataset]
    show_experiment_configuration(args, dataset_info, model_type)
    # Convert the loaded dataset to a stream
    print_colored("[Loading the dataset...]")
    # TODO - conditionally enable validation stream
    train_stream = dataset_info.dataset_loader()[0]
    if args.num_shots is not None:
        train_stream = subsample_stream(train_stream, args.num_shots)
    # Init the resource & Build the tuning config from our dataset/arg info
    print_colored("[Loading the base model resource...]")
    base_model = model_type.bootstrap(args.model_name, tokenizer_name=args.model_name)
    tuning_config = build_tuning_config(args, dataset_info)
    # Then actually train the model & save it
    print_colored("[Starting the training...]")
    model = PeftPromptTuning.train(
        base_model,
        train_stream,
        tuning_config,
        val_stream=None,
        max_source_length=args.max_source_length,
        max_target_length=args.max_target_length,
        tuning_type=args.tuning_type,
        num_epochs=args.num_epochs,
        learning_rate=args.learning_rate,
        batch_size=args.batch_size,
        verbalizer=dataset_info.verbalizer,
        silence_progress_bars=not args.verbose,
        accumulate_steps=args.accumulate_steps,
        torch_dtype=args.torch_dtype,
    )
    model.save(args.output_dir, save_base_model=not args.prompt_only)
    print_colored("[Training Complete]")



"""This script loads and runs a sample PEFT model as a caikit module using
the TGIS backend.

In a nutshell, this does the following:
- Check if `text-generation-launcher` is defined; if it doesn't, assume we have a Docker image
  that is (currently hardcoded to be) able to run TGIS & expose the proper ports, and patch a
  wrapper script around it onto our path so that the TGIS backend falls back to leveraging it
- Load the model through caikit
- Run an inference text generation request and dump the (garbage) output back to the console
"""
# Standard
from shutil import which
import os
import subprocess
import sys

# First Party
from caikit.core.module_backend_config import _CONFIGURED_BACKENDS, configure
from caikit_tgis_backend import TGISBackend
import alog
import caikit

# Local
import caikit_nlp

alog.configure("debug4")

PREFIX_PATH = "prompt_prefixes"

has_text_gen = which("text-generation-launcher")
if not which("text-generation-launcher"):
    print("Text generation server command not found; using Docker override")
    this_dir = os.path.dirname(os.path.abspath(__file__))
    os.environ["PATH"] += ":" + this_dir
    assert (
        which("text-generation-launcher") is not None
    ), "Text generation script not found!"

# Configure caikit to prioritize TGIS backend
_CONFIGURED_BACKENDS.clear()
# load_timeout: 320
#           grpc_port: null
#           http_port: 3001
#           health_poll_delay: 1.0
caikit.configure(
    config_dict={"module_backends": {"priority": [TGISBackend.backend_type]}}
)  # should not be necessary but just in case
configure()  # backend configure

# Load with TGIS backend
prefix_model_path = os.path.join(PREFIX_PATH, "sample_prompt")
my_model = caikit.load(prefix_model_path)
sample_text = "@TommyHilfiger Dramatic shopping exp. ordered 6 jeans same size (30/32) 2 fits / 2 too large / 2 too slim : same brand &gt; different sizing"
sample_output = my_model.run(sample_text)

print("---------- Model result ----------")
print(sample_output)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Caikit prompt tuning library
"""
# Standard
import os

# First Party
from caikit.core.model_manager import *

# Import the model management semantics from the core
import caikit

# Local
# Import subpackages
from . import config, data_model, model_management
from .config import *
from .data_model import *
from .modules import *
from .resources import *
from .version import __version__, __version_tuple__

# Configure the library with library-specific configuration file
CONFIG_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "config", "config.yml")
)

caikit.configure(CONFIG_PATH)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Huggingface auto causal LM resource type
"""
# Standard
from collections.abc import Mapping
from typing import Dict, List, Union

# Third Party
from torch.utils.data import IterableDataset
from transformers import (
    AutoModelForSeq2SeqLM,
    DataCollatorForSeq2Seq,
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
)
from transformers.models.auto import modeling_auto

# First Party
from caikit.core.exceptions import error_handler
from caikit.core.modules import module
import alog

# Local
from ...data_model import GenerationTrainRecord, PromptOutputModelType
from ...toolkit.trainer_utils import log_step
from ...toolkit.verbalizer_utils import render_verbalizer
from .base import PretrainedModelBase

log = alog.use_channel("HFRBAS")
error = error_handler.get(log)

IGNORE_ID = -100


class LoggingTrainer(Seq2SeqTrainer):
    def log(self, logs: Dict[str, float]) -> None:
        """
        Log `logs` on the various objects watching training.

        Subclass and override this method to inject custom behavior.

        Args:
            logs (`Dict[str, float]`):
                The values to log.
        """
        self.state = log_step(self.state, logs)
        self.control = self.callback_handler.on_log(
            self.args, self.state, self.control, logs
        )


@module(
    id="6759e891-287b-405b-bd8b-54a4a4d51c25",
    name="HF Transformers Auto Seq2Seq LM",
    version="0.1.0",
)
class HFAutoSeq2SeqLM(PretrainedModelBase):
    """This resource (module) wraps a handle to a Huggingface
    AutoModelForSeq2SeqLM
    """

    MODEL_TYPE = AutoModelForSeq2SeqLM
    SUPPORTED_MODEL_TYPES = modeling_auto.MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING_NAMES
    TASK_TYPE = "SEQ_2_SEQ_LM"
    PROMPT_OUTPUT_TYPES = [PromptOutputModelType.ENCODER]
    MAX_NUM_TRANSFORMERS = 2

    @classmethod
    def get_num_transformers_submodules(
        cls, output_model_types: List[PromptOutputModelType]
    ):
        """Return number of applicable transformer submodules"""
        num_transformer_submodules = 0
        if PromptOutputModelType.ENCODER in output_model_types:
            num_transformer_submodules += 1
        if PromptOutputModelType.DECODER in output_model_types:
            num_transformer_submodules += 1
        error.value_check(
            "<NLP71505742E>", 0 < num_transformer_submodules <= cls.MAX_NUM_TRANSFORMERS
        )
        return num_transformer_submodules

    def get_trainer(
        self,
        train_dataset: IterableDataset,
        eval_dataset: Union[IterableDataset, None] = None,
        optimizers=(None, None),
        **kwargs
    ):
        """
        Args:
            *kwargs: arguments supported by HF Seq2SeqTrainingArguments:
            https://huggingface.co/docs/transformers/v4.30.0/en/main_classes/trainer#transformers.Seq2SeqTrainingArguments

        NOTE: following parameters are not supported currently:
            1. model_init
            2. compute_metrics
            3. callbacks
            4. preprocess_logits_for_metrics
        """

        # NOTE: predict_with_generate is incompatible with fsdp
        training_args = Seq2SeqTrainingArguments(**kwargs)

        # pylint: disable=duplicate-code
        # TODO: Fetch DataCollator either from property of this
        # class or fetch it as an argument.
        data_collator = self._get_data_collator(**kwargs)

        trainer_arguments = {
            "train_dataset": train_dataset,
            "data_collator": data_collator,
            "optimizers": optimizers,
            "eval_dataset": eval_dataset,
            # "generation_max_length": max_target_length,
        }

        return LoggingTrainer(self._model, training_args, **trainer_arguments)

    def _get_data_collator(self, **kwargs):
        """Function to return appropriate data collator based on resource.

        This implementation uses DataCollatorForSeq2Seq

        Args:
            **kwargs:
                All the keyword arguments passed to this function
                will get filtered out to appropriate ones that are
                applicable to implemented data collator.
        Returns:
            transformers.DataCollator
        """

        applicable_args = ["max_length", "pad_to_multiple_of"]
        collator_kwargs = {key: kwargs[key] for key in applicable_args if key in kwargs}

        return DataCollatorForSeq2Seq(
            tokenizer=self._tokenizer, model=self._model, **collator_kwargs
        )

    @classmethod
    def tokenize_function(
        cls,
        example: Union[GenerationTrainRecord, Mapping],
        tokenizer: "AutoTokenizer",
        max_source_length: int,
        max_target_length: int,
        verbalizer: Union[None, str] = None,
        task_ids: Union[None, int] = None,
    ) -> "BatchEncoding":
        """Tokenization function to be used for seq2seq training; this function consumes a
        GenerationTrainRecord object and applies the verbalizer to it followed by
        the model tokenizer. Finally, we postprocess by ignoring pad tokens in the label IDs.

        Args:
            example: Union[GenerationTrainRecord, Mapping]
                Training data model object to convert a form we can learn on.

        Returns:
            transformers.tokenization_utils_base.BatchEncoding
                encoded tokenization output corresponding to the input example.
        """
        source, target = cls.decompose_example_io(example)
        source = (
            source if verbalizer is None else render_verbalizer(verbalizer, example)
        )

        model_inputs = tokenizer(
            source,
            max_length=max_source_length,
            padding="max_length",
            truncation=True,
        )
        labels = tokenizer(
            target,
            max_length=max_target_length,
            padding="max_length",
            truncation=True,
        )

        labels = labels["input_ids"]

        labels = list(
            map(lambda x: IGNORE_ID if x == tokenizer.pad_token_id else x, labels)
        )
        model_inputs["labels"] = labels
        if task_ids is not None:
            model_inputs["task_ids"] = task_ids

        return model_inputs



"""Given a trained model, which was presumably created by run_peft_tuning.py,
load it and run evaluation.
"""
# Standard
import argparse
import json
import pathlib

# Third Party
from tqdm import tqdm
from utils import (
    SUPPORTED_DATASETS,
    SUPPORTED_METRICS,
    configure_random_seed_and_logging,
    get_wrapped_evaluate_metric,
    is_float,
    kill_tgis_container_if_exists,
    load_model,
    print_colored,
    string_to_float,
)

# Local
from caikit_nlp.toolkit.verbalizer_utils import render_verbalizer


def parse_args() -> argparse.Namespace:
    """Parse & validate command line arguments.

    Returns:
        argparse.Namespace
            Parsed arguments to be leveraged model evaluation.
    """
    parser = argparse.ArgumentParser(
        description="Evaluate a text generation model.",
    )
    # TODO - Patch text-generation-launcher model var so that we can't mount the wrong model
    parser.add_argument(
        "--tgis",
        help="If enabled, runs inference through TGIS instead the local .run().",
        action="store_true",
    )
    parser.add_argument(
        "--model_path",
        help="Model to be loaded from disk.",
        type=pathlib.Path,
        required=True,
    )
    parser.add_argument(
        "--dataset",
        help="Dataset to use to train prompt vectors. Options: {}".format(
            list(SUPPORTED_DATASETS.keys())
        ),
        default="twitter_complaints",
    )
    parser.add_argument(
        "--metrics",
        help="Metrics to calculate in space delimited list",
        default=["accuracy"],
        nargs="*",
        choices=list(SUPPORTED_METRICS.keys()),
    )
    parser.add_argument(
        "--preds_file",
        help="JSON file to dump raw source / target texts to.",
        default="model_preds.json",
    )
    parser.add_argument(
        "--max_new_tokens",
        help="Maximum number of new tokens to be generated",
        type=int,
        default=20,
    )
    parser.add_argument(
        "--truncate_input_tokens",
        help="Number of allowed input tokens (no truncation=0)",
        type=int,
        default=0,
    )
    args = parser.parse_args()
    return args


def get_model_preds_and_references(
    model, validation_stream, max_new_tokens, truncate_input_tokens
):
    """Given a model & a validation stream, run the model against every example in the validation
    stream and compare the outputs to the target/output sequence.

    Args:
        model
            Peft Model to be evaluated (may leverage different backends).
        validation_stream: DataStream[GenerationTrainRecord]
            Validation stream with labeled targets that we want to compare to our model's
            predictions.
        max_new_tokens: int
            Max number of new tokens to be generated, i.e., output limit
        truncate_input_tokens: int
            Number of allowed input tokens, i.e., input limit

    Returns:
        Tuple(List)
            Tuple of 2 lists; the model predictions and the expected output sequences.
    """
    model_preds = []
    targets = []

    for datum in tqdm(validation_stream):
        # Local .run() currently prepends the input text to the generated string;
        # Ensure that we're just splitting the first predicted token & beyond.
        raw_model_text = model.run(
            datum.input,
            max_new_tokens=max_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
        ).generated_text
        parse_pred_text = raw_model_text.split(datum.input)[-1].strip()
        model_preds.append(parse_pred_text)
        targets.append(datum.output)
    return (
        model_preds,
        targets,
    )


def export_model_preds(preds_file, predictions, validation_stream, verbalizer):
    """Exports a JSON file containing a list of objects, where every object contains:
        - source: str - Source string used for generation.
        - target: str - Ground truth target label used for generation.
        - verbalized_source: str - Source string after model verbalization
        - predicted_target: str - Predicted model target.

    Args:
        preds_file: str
            Path on disk to JSON file to be written.
        predictions: List
            Model prediction list, where each predicted text excludes source text as a prefix.
        validation_stream: DataStream
            Datastream object of GenerationTrainRecord objects used for validation against a model
            to generate predictions.
        verbalizer: str
            Model verbalizer used for generating target predictions.
    """
    pred_objs = []
    for pred, record in zip(predictions, validation_stream):
        res = {
            "source": record.input,
            "target": record.output,
            "predicted_target": pred,
        }
        if verbalizer is not None:
            res["verbalized_source"] = render_verbalizer(verbalizer, record)
        pred_objs.append(res)

    with open(preds_file, "w") as jfile:
        json.dump(pred_objs, jfile, indent=4, sort_keys=True)


if __name__ == "__main__":
    configure_random_seed_and_logging()
    args = parse_args()
    metric_funcs = [SUPPORTED_METRICS[metric] for metric in args.metrics]
    print_colored("Metrics to be calculated: {}".format(args.metrics))

    # Load the model; this can be a local model, or a distributed TGIS instance
    print_colored("Loading the model...")
    model = load_model(args.tgis, str(args.model_path))
    # Load the validation stream with marked target sequences
    print_colored("Grabbing validation data...")
    dataset_info = SUPPORTED_DATASETS[args.dataset]
    validation_stream = dataset_info.dataset_loader()[1]
    if validation_stream is None:
        raise ValueError(
            "Selected dataset does not have a validation dataset available!"
        )

    # Run the data through the model; save the predictions & references
    print_colored("Getting model predictions...")
    predictions, references = get_model_preds_and_references(
        model, validation_stream, args.max_new_tokens, args.truncate_input_tokens
    )
    print_colored(
        "Exporting model preds, source, verbalized source, and ground truth targets to {}".format(
            args.preds_file
        )
    )
    export_model_preds(
        args.preds_file,
        predictions,
        validation_stream,
        getattr(model, "verbalizer", None),
    )

    for metric_func in metric_funcs:
        metric_res = metric_func(predictions=predictions, references=references)
        print_colored(metric_res)
    # If we started a TGIS instance, kill it; otherwise, leave our container alone.
    # TODO: This will still looks for containers to kill, even if you're running TGIS
    # outside of a container through text-generation-server directly. For now, we are
    # always running TGIS in a container, so it's ok; the worst that will happen is
    # you'll kill somebody else's container.
    if args.tgis:
        print_colored("Killing containerized TGIS instance...")
        kill_tgis_container_if_exists()



"""Tests for regex sentence splitter
"""
# Standard
import os
import tempfile

# First Party
from caikit.interfaces.nlp.data_model import TokenizationResults

# Local
from caikit_nlp.modules.tokenization.regex_sentence_splitter import (
    RegexSentenceSplitter,
)

## Setup ########################################################################

# Regex sentence splitter model for reusability across tests
REGEX_STR = "[^.!?\s][^.!?\n]*(?:[.!?](?!['\"]?\s|$)[^.!?]*)*[.!?]?['\"]?(?=\s|$)"
SENTENCE_TOKENIZER = RegexSentenceSplitter.bootstrap(REGEX_STR)
DOCUMENT = "What he told me before, I have it in my heart. I am tired of fighting."

## Tests ########################################################################


def test_bootstrap_and_run():
    """Check if we can bootstrap and run regex sentence splitter"""
    tokenization_result = SENTENCE_TOKENIZER.run(DOCUMENT)
    assert isinstance(tokenization_result, TokenizationResults)
    assert len(tokenization_result.results) == 2


def test_save_load_and_run_model():
    """Check if we can run a saved model successfully"""
    with tempfile.TemporaryDirectory() as model_dir:
        SENTENCE_TOKENIZER.save(model_dir)
        assert os.path.exists(os.path.join(model_dir, "config.yml"))

        new_splitter = RegexSentenceSplitter.load(model_dir)
        tokenization_result = new_splitter.run(DOCUMENT)
        assert isinstance(tokenization_result, TokenizationResults)
        assert len(tokenization_result.results) == 2



"""Tests for verbalizer substitution utils.
"""
# Third Party
import pytest

# Local
from caikit_nlp.toolkit.verbalizer_utils import is_valid_verbalizer, render_verbalizer
import caikit_nlp

SAMPLE_DM = caikit_nlp.data_model.GenerationTrainRecord(
    input="my input text", output="my output text"
)
SAMPLE_DICT = SAMPLE_DM.to_dict()

### Happy path rendering cases
def test_is_valid_verbalizer():
    """Ensure that when we have happy verbalizers, it's easy to check."""
    assert is_valid_verbalizer("{{input}}") == True
    assert is_valid_verbalizer("Input: {{input}}") == True
    assert is_valid_verbalizer("Input: {{input}} Output: {{output}}") == True


def test_raw_verbalizer_is_preserved():
    """Ensure that if no placeholders are provided, the raw string is returned."""
    # NOTE: you would never want to do this because it means your input string is hardcoded!
    # However, we the responsibility of checking this up to the verbalizer validator so that
    # we can validate up front, rather than checking lazily at train time etc.
    verbalizer_template = "text: foo label: bar"
    expected_render_result = verbalizer_template
    # Check that we get the same behavior from both the DM / Dict based verbalizer replacement
    assert render_verbalizer(verbalizer_template, SAMPLE_DM) == expected_render_result
    assert render_verbalizer(verbalizer_template, SAMPLE_DICT) == expected_render_result


def test_verbalizer_substitution():
    """Ensure that if placeholders are used, class attrs are rendered into placeholders."""
    verbalizer_template = "text: {{input}} label: {{output}}"
    expected_render_result = "text: {} label: {}".format(
        SAMPLE_DM.input, SAMPLE_DM.output
    )
    assert render_verbalizer(verbalizer_template, SAMPLE_DM) == expected_render_result
    assert render_verbalizer(verbalizer_template, SAMPLE_DICT) == expected_render_result


def test_invalid_value_verbalizer_substitution():
    """Ensure that if we try to render a placeholder that is a bad property name, we skip it."""
    verbalizer_template = "text: {{foo bar}}"
    expected_render_result = verbalizer_template
    assert render_verbalizer(verbalizer_template, SAMPLE_DM) == expected_render_result
    assert render_verbalizer(verbalizer_template, SAMPLE_DICT) == expected_render_result


def test_empty_verbalizer_substitution():
    """Ensure that if we try to render an empty placeholder, we skip it."""
    verbalizer_template = "text: {{}} label: {{}}"
    expected_render_result = verbalizer_template
    assert render_verbalizer(verbalizer_template, SAMPLE_DM) == expected_render_result
    assert render_verbalizer(verbalizer_template, SAMPLE_DICT) == expected_render_result


### sad path rendering cases
def test_is_invalid_verbalizer():
    """Ensure that when we have happy verbalizers, it's easy to check."""
    assert is_valid_verbalizer(100) == False
    assert is_valid_verbalizer("") == False
    assert is_valid_verbalizer("source") == False
    assert is_valid_verbalizer("{{this is not a valid placeholder}}") == False


def test_invalid_attribute_verbalizer_substitution():
    """Ensure that if we try to render a placeholder that doesn't exist on our DM, we fail."""
    verbalizer_template = "text: {{sadness}}"
    with pytest.raises(AttributeError):
        render_verbalizer(verbalizer_template, SAMPLE_DM)


def test_invalid_key_verbalizer_substitution():
    """Ensure that if we try to render a placeholder that doesn't exist on our dict, we fail."""
    verbalizer_template = "text: {{sadness}}"
    with pytest.raises(KeyError):
        render_verbalizer(verbalizer_template, SAMPLE_DICT)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Callable, Dict, List, Optional, Tuple, Type, Union
import json
import os

# Third Party
from torch.utils.data import IterableDataset
from transformers import (
    AutoTokenizer,
    DataCollatorWithPadding,
    PreTrainedTokenizerBase,
    Trainer,
    TrainingArguments,
)
from transformers.models.auto.auto_factory import _BaseAutoModelClass
import torch

# First Party
from caikit import get_config
from caikit.core.data_model import DataStream
from caikit.core.exceptions import error_handler
from caikit.core.modules import ModuleBase, ModuleConfig, ModuleSaver
import alog

# Local
from ...data_model import GenerationTrainRecord, PromptOutputModelType
from ...toolkit.data_type_utils import get_torch_dtype, str_to_torch_dtype
from ...toolkit.trainer_utils import log_step

log = alog.use_channel("HFRBAS")
error = error_handler.get(log)


class LoggingTrainer(Trainer):
    def log(self, logs: Dict[str, float]) -> None:
        """
        Log `logs` on the various objects watching training.

        Subclass and override this method to inject custom behavior.

        Args:
            logs (`Dict[str, float]`):
                The values to log.
        """
        self.state = log_step(self.state, logs)
        self.control = self.callback_handler.on_log(
            self.args, self.state, self.control, logs
        )


class PretrainedModelBase(ABC, ModuleBase):
    """Common abstractions and requirements for pretrained model resources"""

    _TOK_ARTIFACTS_CONFIG_KEY = "tokenizer_artifacts"
    _MODEL_ARTIFACTS_CONFIG_KEY = "model_artifacts"
    _LEFT_PAD_MODEL_TYPES = ("gpt", "opt", "bloom")

    @classmethod
    @property
    def REQUIRES_TOKEN_UNWRAPPING(cls) -> str:
        """Most models don't need token unwrapping from their tokenizer closures"""
        return False

    ## Abstract Interface ######################################################

    @classmethod
    @property
    @abstractmethod
    def MODEL_TYPE(cls) -> Type[_BaseAutoModelClass]:
        """All classes must have a class property declaring the type of HF model
        they support
        """

    @classmethod
    @property
    @abstractmethod
    def TASK_TYPE(cls) -> str:
        """All classes must have indicate the PEFT task type that they use"""

    @classmethod
    @property
    @abstractmethod
    def SUPPORTED_MODEL_TYPES(cls) -> str:
        """All classes must indicate the model types supported by the resource"""

    ## Shared Implementation ###################################################
    def __init__(
        self,
        tokenizer: AutoTokenizer,
        model: _BaseAutoModelClass,
        model_name: str,
        torch_dtype: torch.dtype,
    ):
        """Initialize with an in-memory handle to a model"""
        super().__init__()
        self._tokenizer = tokenizer
        self._model = model
        self._model_name = model_name
        self._torch_dtype = torch_dtype

    @property
    def model(self) -> _BaseAutoModelClass:
        """Get access to the underlying causal LM"""
        return self._model

    @property
    def tokenizer(self) -> AutoTokenizer:
        """Get access to the underlying tokenizer"""
        return self._tokenizer

    def get_config(self):
        """Function to return model config from transformer model"""
        # This funky json.load step is just to make sure any non-string
        # keys gets converted to string automatically otherwise
        # it will throw error when we try to save them using ModuleSaver.
        # if it was not for this, we could have just used self._module.config.to_dict()
        return json.loads(self.model.config.to_json_string())

    @classmethod
    def bootstrap(
        cls,
        model_name: str,
        tokenizer_name: Optional[str] = None,
        padding_side: Optional[str] = None,
        torch_dtype: Optional[torch.dtype] = None,
        **kwargs,
    ) -> "HFAutoSequenceClassifier":
        """Bootstrap from a huggingface model

        This method takes the raw HF model name / path and bootstraps the
        tokenizer in memory. If downloads are enabled, the model name can be any
        model name from HF, otherwise it must be a path to an on-disk model or
        the name of a pre-cached model.

        Args:
            model_name (str)
                The name/path of the HF sequence classifier model
            tokenizer_name (Optional[Union[str, PreTrainedTokenizerBase]])
                The name/path of the HF tokenizer model (matches model_name if
                not given) or an instance of a loaded tokenizer.
                NOTE: If a loaded tokenizer is provided, and it doesn't have
                a pad token ID, the pad token ID will be set to the EOS token ID.
            padding_side (Optional[str])
                The padding side for the tokenizer. Found by convention if not
                given. This value is only used if a tokenizer needs to be loaded.
            torch_dtype: (Optional[Union[torch.dtype, str]])
                Data type to load the model as; if no value is provided, we pull
                torch_dtype from config.
            **kwargs
                Additional keyword args to pass to from_pretrained
                (e.g. return_dict=True)

        Returns:
            model HFAutoSequenceClassifier
                The loaded resource model
        """

        torch_dtype = get_torch_dtype(torch_dtype)

        # Check if we passed the tokenizer directly; for now, we keep
        # the arg name tokenizer_name for compatibility reasons
        if isinstance(tokenizer_name, PreTrainedTokenizerBase):
            log.debug("Bootstrapping with in-memory tokenizer")
            tokenizer = tokenizer_name

        else:
            if tokenizer_name is None:
                tokenizer_name = model_name
                log.info("Loading tokenizer from model directory")

            if not os.path.isdir(tokenizer_name) and tokenizer_name != model_name:
                log.warning(
                    "Bootstrapping with mismatched tokenizer (%s) / model (%s)",
                    tokenizer_name,
                    model_name,
                )

            # Figure out the right padding side based on the name of the HF model
            # NOTE: This matches models whose name includes the left-pad types as a
            #   substring and not just as an exact match.
            if padding_side is None:
                padding_side = (
                    "left"
                    if any(k in model_name for k in cls._LEFT_PAD_MODEL_TYPES)
                    else "right"
                )

            with alog.ContextTimer(log.info, "Tokenizer loaded in "):
                # Load the tokenizer and set up the pad token if needed
                tokenizer = AutoTokenizer.from_pretrained(
                    tokenizer_name,
                    local_files_only=not get_config().allow_downloads,
                    padding_side=padding_side,
                    # We can't disable use_fast otherwise unit test fails
                    # use_fast=False,
                )

        if tokenizer.pad_token_id is None:
            tokenizer.pad_token_id = tokenizer.eos_token_id

        with alog.ContextTimer(log.info, f"Model {model_name} loaded in "):
            # Load the model
            model = cls.MODEL_TYPE.from_pretrained(
                model_name,
                local_files_only=not get_config().allow_downloads,
                torch_dtype=torch_dtype,
                **kwargs,
            )
        log.debug4("Model Details: %s", model)

        # Create the class instance
        inst_model_name = os.path.split(model_name)[-1]
        return cls(
            tokenizer=tokenizer,
            model=model,
            model_name=inst_model_name,
            torch_dtype=torch_dtype,
        )

    @classmethod
    def load(cls, model_path: str) -> Type["PretrainedModelBase"]:
        """Load from a saved resource model"""
        model_path = os.path.abspath(model_path)
        config = ModuleConfig.load(model_path)
        tok_abs_path = os.path.join(model_path, config[cls._TOK_ARTIFACTS_CONFIG_KEY])
        model_abs_path = os.path.join(
            model_path, config[cls._MODEL_ARTIFACTS_CONFIG_KEY]
        )
        error.dir_check("<NLP12813455E>", tok_abs_path)
        error.dir_check("<NLP12813443E>", model_abs_path)
        res = cls.bootstrap(
            tokenizer_name=tok_abs_path,
            model_name=model_abs_path,
            padding_side=config.padding_side,
            torch_dtype=str_to_torch_dtype(config.torch_dtype),
        )
        return res

    def save(
        self, model_path: str, tokenizer_dirname: str = "", base_model_dirname: str = ""
    ):
        """Save the in-memory model to the given path"""
        saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        tok_rel_path, tok_abs_path = saver.add_dir(tokenizer_dirname)
        model_rel_path, model_abs_path = saver.add_dir(base_model_dirname)
        with saver:
            saver.update_config(
                {
                    self._TOK_ARTIFACTS_CONFIG_KEY: tok_rel_path,
                    self._MODEL_ARTIFACTS_CONFIG_KEY: model_rel_path,
                    "padding_side": self.tokenizer.padding_side,
                    "model_name": self._model_name,
                    # Grab the torch property for the dtype so that we can rebuild from a str.
                    "torch_dtype": str(self._torch_dtype).rsplit(".", maxsplit=1)[-1],
                }
            )
            self.tokenizer.save_pretrained(tok_abs_path)
            self.model.save_pretrained(model_abs_path)

    def get_trainer(
        self,
        train_dataset: IterableDataset,
        eval_dataset: Union[IterableDataset, None] = None,
        optimizers=(None, None),
        **kwargs,
    ):
        """
        Args:
            **kwargs: arguments supported by HF TrainingArguments:
            https://huggingface.co/docs/transformers/v4.30.0/en/main_classes/trainer#transformers.TrainingArguments

        NOTE: following parameters are not supported currently:
            1. model_init
            2. compute_metrics
            3. callbacks
            4. preprocess_logits_for_metrics
        """

        training_args = TrainingArguments(**kwargs)

        data_collator = self._get_data_collator(**kwargs)

        trainer_arguments = {
            "train_dataset": train_dataset,
            "data_collator": data_collator,
            "optimizers": optimizers,
            "eval_dataset": eval_dataset,
        }

        return LoggingTrainer(self._model, training_args, **trainer_arguments)

    def _get_data_collator(self, **kwargs):
        """Function to return appropriate data collator based on resource.

        The default implementation of the base resource uses
        DataCollatorWithPadding which will dynamically pad the inputs received.

        Args:
            **kwargs:
                All the keyword arguments passed to this function
                will get filtered out to appropriate ones that are
                applicable to implemented data collator.
        Returns:
            transformers.DataCollator
        """

        applicable_args = ["max_length", "pad_to_multiple_of"]
        collator_kwargs = {key: kwargs[key] for key in applicable_args if key in kwargs}

        return DataCollatorWithPadding(
            tokenizer=self._tokenizer, padding=True, **collator_kwargs
        )

    # pylint: disable=unused-argument
    @classmethod
    def get_num_transformers_submodules(
        cls, output_model_types: List[PromptOutputModelType]
    ):
        """Return number of applicable transformer submodules"""
        return 1

    @staticmethod
    def decompose_example_io(example: Union[GenerationTrainRecord, Mapping]):
        """Given an example, which might be a number of supported types,
        extract the input / output texts. Depending on the manner in which
        the sample is being leveraged, this might be a raw data model object,
        a dict, or some other mappable, e.g., a HF dataset LazyRow.

        args:
            example: Union[GenerationTrainRecord, Mapping]
                Objects whose input / output we want to retrieve.

        Returns:
            Tuple[str, str]
                Input & Output strings.
        """
        if isinstance(example, GenerationTrainRecord):
            return example.input, example.output
        # TODO: probably a good idea to add some error handling here;
        # For now, we don't since situations in which we call this
        # internally should generally enforce this as true, e.g.,
        # hf datasets created out of data model objects.
        return example["input"], example["output"]

    @classmethod
    def build_task_tokenize_closure(cls, *args, **kwargs) -> Tuple[Callable, bool]:
        """Builds tokenizer closure which can be mapped over train streams to process
        data which can then be easily passed to a DataLoader for different model types.
        This is largely for convenience if we want a closure that can be applied
        without having to carry around other parameters.
        """

        def tokenize_wrapper(example: GenerationTrainRecord):
            return cls.tokenize_function(example, *args, **kwargs)

        return (tokenize_wrapper, cls.REQUIRES_TOKEN_UNWRAPPING)

    @classmethod
    @abstractmethod
    def tokenize_function(
        cls,
        example: Union[GenerationTrainRecord, Mapping],
        tokenizer: "AutoTokenizer",
        max_source_length: int,
        max_target_length: int,
        verbalizer: Union[None, str] = None,
        task_ids: Union[None, int] = None,
    ) -> Union["BatchEncoding", DataStream["BatchEncoding"]]:
        """Tokenizes a generation training record.

        Args:
            Union[GenerationTrainRecord, Mapping]
                Example data model object / mapping to be tokenized.
            tokenizer: AutoTokenizer
                Model tokenizer to be used in preprocessing, i.e., when we iterate over our data.
            max_source_length: int
                Max length of sequences being considered.
            max_target_length: int
                Max length of target sequences being predicted.
            verbalizer: str
                Verbalizer template to be used for formatting data. This template may use brackets
                to indicate where fields from the data model TrainGenerationRecord must be rendered.
                If no verbalizer is provided, the source text is used as the rendered result.
            task_ids: Union[None, int]
                Task id corresponding particular task for multi-task prompt tuning.
                NOTE: Only required for MPT (Multi-task prompt tuning)
                Default: None

        Returns:
            BatchEncoding | DataStream[BatchEncoding]
                encoded tokenization output corresponding to the input example.
        """



# First Party
from caikit.interfaces.nlp.data_model import FinishReason, GeneratedTextResult

generated_response = GeneratedTextResult(
    generated_text="foo bar",
    generated_tokens=2,
    finish_reason=FinishReason.STOP_SEQUENCE,
)

# Add an example of one of each new data model types that you've defined within your extension
# to the list below. These samples are used for verifying that your object is well-aligned with
# with caikit serialization interfaces. They will only be used if your extension enables
# protobuf serialization in its config.
#
# NOTE: You do not need to add any samples from other extensions or the caikit library; only
# new types explicitly created in your extension data model.
data_model_samples = [
    generated_response,
]



# Third Party
import pytest

# First Party
from caikit.core.data_model.producer import ProducerId
from caikit.interfaces.nlp.data_model import GeneratedTextResult

# Local
from caikit_nlp.toolkit.text_generation.model_run_utils import generate_text_func
from tests.fixtures import (
    causal_lm_dummy_model,
    causal_lm_train_kwargs,
    seq2seq_lm_dummy_model,
    seq2seq_lm_train_kwargs,
)


@pytest.mark.parametrize(
    "model_fixture", ["seq2seq_lm_dummy_model", "causal_lm_dummy_model"]
)
@pytest.mark.parametrize(
    "serialization_method,expected_type",
    [
        ("to_dict", dict),
        ("to_json", str),
        ("to_proto", GeneratedTextResult._proto_class),
    ],
)
def test_generate_text_func_serialization_json(
    request,
    model_fixture,
    serialization_method,
    expected_type,
):
    model = request.getfixturevalue(model_fixture)
    generated_text = generate_text_func(
        model=model.model,
        tokenizer=model.tokenizer,
        producer_id=ProducerId("TextGeneration", "0.1.0"),
        eos_token="<\n>",
        text="What is the boiling point of liquid Nitrogen?",
    )

    serialized = getattr(generated_text, serialization_method)()
    assert isinstance(serialized, expected_type)


@pytest.mark.parametrize("causal_model_fixture", ["causal_lm_dummy_model"])
def test_generate_text_func_preserve_input_causal_lm(request, causal_model_fixture):
    """For Causal LM task types, setting preserve_inout_text to True
    will result in input text in model prediction. Setting to False will
    strip the input text from model prediction.
    """
    input_text = "What is the boiling point of liquid Nitrogen?"
    causal_model = request.getfixturevalue(causal_model_fixture)
    # assert type(causal_model.model) == False
    generated_text = generate_text_func(
        model=causal_model.model,
        tokenizer=causal_model.tokenizer,
        producer_id=ProducerId("TextGeneration", "0.1.0"),
        eos_token="<\n>",
        text=input_text,
        preserve_input_text=True,
        task_type="CAUSAL_LM",
    )
    assert input_text in generated_text.generated_text
    generated_text = generate_text_func(
        model=causal_model.model,
        tokenizer=causal_model.tokenizer,
        producer_id=ProducerId("TextGeneration", "0.1.0"),
        eos_token="<\n>",
        text=input_text,
        preserve_input_text=False,
        task_type="CAUSAL_LM",
    )
    assert input_text not in generated_text.generated_text


@pytest.mark.parametrize("seq_model_fixture", ["seq2seq_lm_dummy_model"])
def test_generate_text_func_preserve_input(request, seq_model_fixture):
    """For Seq2Seq LM task types, setting preserve_inout_text to True
    or False should not change predictions.
    """
    input_text = "What is the boiling point of liquid Nitrogen?"
    seq_model = request.getfixturevalue(seq_model_fixture)
    # assert type(causal_model.model) == False
    generated_text = generate_text_func(
        model=seq_model.model,
        tokenizer=seq_model.tokenizer,
        producer_id=ProducerId("TextGeneration", "0.1.0"),
        eos_token="<\n>",
        text=input_text,
        preserve_input_text=True,
        task_type="SEQ_2_SEQ_LM",
    )
    before_pred = generated_text.generated_text
    generated_text = generate_text_func(
        model=seq_model.model,
        tokenizer=seq_model.tokenizer,
        producer_id=ProducerId("TextGeneration", "0.1.0"),
        eos_token="<\n>",
        text=input_text,
        preserve_input_text=False,
        task_type="SEQ_2_SEQ_LM",
    )
    after_pred = generated_text.generated_text
    assert before_pred == after_pred



"""Tests for data type related utils, e.g., for interacting with serialized torch types.
"""
# Third Party
import pytest
import torch

# Local
from caikit_nlp.toolkit.data_type_utils import get_torch_dtype, str_to_torch_dtype

### Tests for converting from strings / types / None -> torch data types


def test_get_torch_dtype_from_str():
    """Ensure that we can parse a data type from a string."""
    assert torch.float32 is get_torch_dtype("float32")


def test_get_torch_dtype_from_dtype():
    """Ensure that if a data type is provided from pytorch, we simply return it."""
    assert torch.float32 is get_torch_dtype(torch.float32)


def test_get_torch_dtype_from_bad_type():
    """Ensure that if a type we can't coerce to a pytorch dtype is given, we get a TypeError."""
    with pytest.raises(TypeError):
        assert torch.float32 is get_torch_dtype(100)


def test_get_torch_dtype_from_bad_str():
    """Ensure that if an invalid type string is given, we get a ValueError."""
    with pytest.raises(ValueError):
        assert torch.float32 is get_torch_dtype("not a valid attr of pytorch")


### Tests for converting from strings -> torch data types
def test_str_to_torch_dtype():
    """Ensure that we can parse a data type from a string."""
    assert str_to_torch_dtype("float32") is torch.float32


def test_str_to_torch_dtype_invalid_attr():
    """Ensure that we raise ValueError if an incorrect type str is provided."""
    with pytest.raises(ValueError):
        str_to_torch_dtype("not a valid attr of pytorch")


def test_str_to_torch_dtype_bad_attr():
    """Ensure that we raise ValueError if a non type property of torch is provided."""
    with pytest.raises(ValueError):
        str_to_torch_dtype("nn")



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Huggingface auto sequence classifier resource type
"""
# Standard
from typing import Callable, Tuple

# Third Party
from transformers import AutoModelForSequenceClassification
from transformers.models.auto import modeling_auto

# First Party
from caikit.core.modules import module

# Local
from .base import PretrainedModelBase


@module(
    id="6759e891-287b-405b-bd8b-54a4a4d51c23",
    name="HF Transformers Auto Sequence Classifier",
    version="0.1.0",
)
class HFAutoSequenceClassifier(PretrainedModelBase):
    """This resource (module) wraps a handle to a huggingface
    AutoModelForSequenceClassification
    """

    MODEL_TYPE = AutoModelForSequenceClassification
    SUPPORTED_MODEL_TYPES = (
        modeling_auto.MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES
    )
    TASK_TYPE = "SEQ_CLS"
    PROMPT_OUTPUT_TYPES = []
    MAX_NUM_TRANSFORMERS = 1

    @classmethod
    def bootstrap(cls, *args, **kwargs) -> "HFAutoSequenceClassifier":
        """Bootstrap from a huggingface model

        See help(PretrainedModelBase)
        """
        return super().bootstrap(*args, return_dict=True, **kwargs)

    @staticmethod
    def tokenize_function(*args, **kwargs) -> Tuple[Callable, bool]:
        raise NotImplementedError(
            "Tokenize func not implemented for sequence classifier"
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Text Embedding Module
=====================

Implements the following tasks:

  1. EmbeddingTask: Returns an embedding from an input text string
  2. EmbeddingsTasks: EmbeddingTask but with a list of inputs producing a list of outputs
  3. SentenceSimilarityTask: Compare one source sentence to a list of sentences
  4. SentenceSimilarityTasks: SentenceSimilarityTask but with a list of source sentences producing
     a list of outputs
  5. RerankTask: Return top_n documents ordered by relevance given a query
  6. RerankTasks: RerankTask but with a list of queries producing a list of outputs

"""

# Local
from .embedding import EmbeddingModule



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Standard
from typing import Any, Dict, Optional, Union
import gc
import json
import os
import tempfile

# Third Party
from datasets import Dataset
from datasets import IterableDataset as TransformersIterableDataset
from transformers import AutoConfig, AutoTokenizer
import torch

# First Party
from caikit import get_config
from caikit.core.data_model import DataStream
from caikit.core.exceptions import error_handler
from caikit.core.modules import ModuleBase, ModuleConfig, ModuleSaver, module
from caikit.interfaces.nlp.data_model import GeneratedTextResult, TokenizationResults
from caikit.interfaces.nlp.tasks import TextGenerationTask, TokenizationTask
import alog

# Local
from ...data_model import GenerationTrainRecord
from ...resources.pretrained_model import (
    HFAutoCausalLM,
    HFAutoSeq2SeqLM,
    PretrainedModelBase,
)
from ...toolkit.data_type_utils import get_torch_dtype, str_to_torch_dtype
from ...toolkit.text_generation.model_run_utils import (
    GENERATE_FUNCTION_ARGS,
    generate_text_func,
)
from ...toolkit.torch_run import get_torch_elastic_launch_config

log = alog.use_channel("TXT_GEN")
error = error_handler.get(log)


TRAINING_LOSS_LOG_FILENAME = "training_logs.jsonl"

# pylint: disable=too-many-lines,too-many-instance-attributes
@module(
    id="f9181353-4ccf-4572-bd1e-f12bcda26792",
    name="Text Generation",
    version="0.1.0",
    tasks=[TextGenerationTask, TokenizationTask],
)
class TextGeneration(ModuleBase):
    """Module to provide text generation capabilities"""

    RANDOM_SEED = 73
    supported_resources = [HFAutoCausalLM, HFAutoSeq2SeqLM]

    # Below list is taken from
    # https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
    allowed_training_args = {
        "weight_decay",
        "adam_beta1",
        "adam_beta2",
        "adam_epsilon",
        "max_grad_norm",
        "lr_scheduler_type",
        "warmup_ratio",
        "warmup_steps",
        "use_ipex",
        "disable_tqdm",
        "label_names",
        "optim",
        "optim_args",
        "group_by_length",
        "dataloader_pin_memory",
        "gradient_checkpointing",
        "full_determinism",
    }

    def __init__(
        self,
        model_name: str,
        model: PretrainedModelBase = None,
        bos_token: Optional[str] = None,
        sep_token: Optional[str] = None,
        eos_token: Optional[str] = None,
        pad_token: Optional[str] = None,
        training_metadata: Union[Dict[str, Any], None] = None,
    ):
        super().__init__()

        error.type_check("<NLP48137045E>", str, allow_none=True, eos_token=eos_token)
        self.model = model
        self.model_name = model_name

        self._bos_token = bos_token
        self._sep_token = sep_token
        self._eos_token = eos_token
        self._pad_token = pad_token
        self.training_metadata = (
            training_metadata if training_metadata is not None else {}
        )

    # pylint: disable=duplicate-code
    def __del__(self):
        del self.model
        gc.collect()
        try:
            torch.cuda.empty_cache()
        except AttributeError:
            pass

    @classmethod
    def bootstrap(cls, base_model_path: str, torch_dtype: str = "float32"):
        """Function to bootstrap a pre-trained transformers model and
        get a caikit text-generation 'model'.

        Args:
            base_model_path: str
                Path to transformers model
                NOTE: Model path needs to contain tokenizer as well
            torch_dtype: str
                Torch data type to be used when loading the model.
                Default: float32
        Returns:
            caikit_nlp.blocks.text_generation.TextGeneration
                Object of TextGeneration class (model)
        """
        # pylint: disable=duplicate-code
        model_config = AutoConfig.from_pretrained(base_model_path)

        resource_type = None
        for resource in cls.supported_resources:
            if model_config.model_type in resource.SUPPORTED_MODEL_TYPES:
                resource_type = resource
                break

        if not resource_type:
            error(
                "<NLP61784225E>",
                "{} model type is not supported currently!".format(
                    model_config.model_type
                ),
            )
        log.debug("Bootstrapping base resource [%s]", base_model_path)
        base_model = resource_type.bootstrap(
            base_model_path,
            tokenizer_name=base_model_path,
            torch_dtype=torch_dtype,
        )
        eos_token = base_model._tokenizer.eos_token or None
        return cls(
            base_model_path,
            base_model,
            eos_token=eos_token,
        )

    @classmethod
    def train(
        cls,
        base_model: str,  # TODO: Union[str, PretrainedModelBase]
        train_stream: DataStream[GenerationTrainRecord],
        torch_dtype: str = None,  # TODO: Optional[Union[torch.dtype, str]]
        max_source_length: int = 256,
        max_target_length: int = 128,
        batch_size: int = 8,
        num_epochs: int = 5,
        accumulate_steps: int = 32,
        random_seed: int = RANDOM_SEED,
        lr: float = 2e-5,
        use_iterable_dataset: bool = True,
        **kwargs,
    ) -> "TextGeneration":
        """
        Fine-tune a CausalLM or Seq2seq text generation model.

        Args:
            base_model:  Union[str, caikit_nlp.resources.pretrained_model.base.PretrainedModelBase]
                Base resource model used for underlying generation.
            train_stream: DataStream[GenerationTrainRecord] or DataStream[ClassificationTrainRecord]
                Data to be used for fine-tuning the generation model.
            torch_dtype: str
                TODO: Optional[Union[torch.dtype, str]]
                Data type to use for training/inference of the underlying text generation model.
                If no value is provided, we pull from torch_dtype in config. If an in memory
                resource is provided which does not match the specified data type, the model
                underpinning the resource will be converted in place to the correct torch dtype.
            max_source_length: int
                Max length of input sequences being considered. Default: 256.
            max_target_length: int
                Max length of target sequences being predicted. Default: 128.
            batch_size: int
                Batch sized to be used for training / evaluation data. Default: 8.
            num_epochs: int
                Number of epochs to tune the model. Default: 20.
            accumulate_steps: int
                Number of steps to use for gradient accumulation. Default: 1.
            lr: float
                Learning rate to be used while tuning model. Default: 2e-5.
            use_iterable_dataset: bool
                Indicates whether or not we should load the full dataset into memory
                NOTE: use True for this option if you are fine tuning a causal LM with
                a large target sequence length unless your dataset is VERY small!
            **kwargs:
                Arguments supported by HF Training Arguments.
                TrainingArguments:
                    https://huggingface.co/docs/transformers/v4.30.0/en/main_classes/trainer#transformers.TrainingArguments
                Seq2SeqTrainingArguments:
                    https://huggingface.co/docs/transformers/v4.30.0/en/main_classes/trainer#transformers.Seq2SeqTrainingArguments
        Returns:
            TextGeneration
                Instance of this class with fine-tuned models.
        """
        error.value_check(
            "<NLP96406893E>", len(train_stream) > 0, "train_stream cannot be empty"
        )

        torch_dtype = get_torch_dtype(torch_dtype)

        # Make training deterministic
        if torch.cuda.is_available():
            torch.backends.cudnn.deterministic = True

        ## NOTE: Below code has been used in couple of places at this point, like in
        # text_generation module. In future, we would want to consolidate this into
        # a base class or a toolkit function
        # pylint: disable=duplicate-code
        resource_type = None

        ## Load base model
        if isinstance(base_model, str):
            model_config = AutoConfig.from_pretrained(base_model)

            for resource in cls.supported_resources:
                if model_config.model_type in resource.SUPPORTED_MODEL_TYPES:
                    resource_type = resource
                    break

            if not resource_type:
                error(
                    "<NLP61784225E>",
                    "{} model type is not supported currently!".format(
                        model_config.model_type
                    ),
                )
            log.debug("Bootstrapping base resource [%s]", base_model)
            base_model = resource_type.bootstrap(base_model, torch_dtype=torch_dtype)

        else:
            # base_model is actually a resource object
            resource_type = type(base_model)

        error.type_check("<NLP03221895E>", PretrainedModelBase, base_model=base_model)

        # TODO; For now, we don't support iterable datasets for causal LM, the reason
        # being that the dynamically padded collator breaks with FSDP/torch run when
        # multiple devices are used. This appears to be because each process fetches
        # part of the batch from the main iterator directly (which does dynamic padding
        # independently per process), which the accelerator data loader wrapper gathers
        # and concatenates to form the batch. I.e., we get concatenation dimension errors.
        #
        # This is a short term workaround, but a more sustainable path forward would be
        # to either force the dataloader workers to 1 without modifying the torchrun config
        # (probably through trainer args), or find a way to form the whole batch on one
        # process and broadcast, although the former is probably better for performance
        # here since things are collected anyway.
        if use_iterable_dataset and isinstance(base_model, HFAutoCausalLM):
            log.warning(
                "<NLP24452569W>",
                "Iterable dataset not supported for CausalLMs; Falling back to normal dataset",
            )
            use_iterable_dataset = False

        ## Generate data loader from stream
        training_dataset: Union[
            Dataset, TransformersIterableDataset
        ] = cls._preprocess_function(
            base_model=base_model,
            train_stream=train_stream,
            tokenizer=base_model.tokenizer,
            max_source_length=max_source_length,
            max_target_length=max_target_length,
            shuffle=True,
            use_iterable_dataset=use_iterable_dataset,
        )

        ### Dtype based processing
        # NOTE: Following is not exhaustive list of all parameters
        # for all dtypes
        if torch_dtype == torch.float16:
            dtype_based_params = {
                "fp16": True,
            }
        elif torch_dtype == torch.bfloat16:
            dtype_based_params = {
                "bf16": True,
            }
        else:
            # default to float32
            dtype_based_params = {}

        ## TODO: Add automatic sharding selection based on number of parameters
        # in base model
        ## TODO: Fetch trainer from resource

        # Filter **training_arguments to only process allowed ones
        filtered_training_arguments = {
            k: v for k, v in kwargs.items() if k in cls.allowed_training_args
        }

        extra_training_args = set(kwargs.keys()).difference(
            filtered_training_arguments.keys()
        )

        if extra_training_args:
            log.warning(
                "<NLP24424909W>",
                f"{extra_training_args} parameter(s) not allowed by \
                    {cls.__name__} currently and will be ignored!",
            )

        processing_configuration = {}

        # Conditionally enable sharding if multiple GPUs available
        if torch.cuda.is_available() and torch.cuda.device_count() > 1:
            processing_configuration = {
                "fsdp": "full_shard offload auto_wrap",
                "fsdp_config": {
                    # NOTE: Every transformers model has `_no_split_modules` property that can be
                    # leveraged to identify the layers to split. This seems to be a decent
                    # "default" behavior unless we want to optimize further. We will start with
                    # this generic approach, since it allows us to handle variety
                    # of models and iterate on it, based on what we encounter.
                    "fsdp_transformer_layer_cls_to_wrap": base_model._model._no_split_modules
                },
            }

        # Open an intermediate checkpoint directory until we've bootstrapped
        # our model or we've early exited (if epochs < 1)
        with tempfile.TemporaryDirectory() as checkpoint_dir:
            training_args = {
                "output_dir": checkpoint_dir,
                "per_device_train_batch_size": batch_size,
                "per_device_eval_batch_size": batch_size,
                "num_train_epochs": num_epochs,
                "seed": random_seed,
                # NOTE: We have disabled evaluation for now
                "do_eval": False,
                "do_train": True,
                "learning_rate": lr,
                "weight_decay": 0.01,
                "save_total_limit": 3,
                "push_to_hub": False,
                "use_cpu": not torch.cuda.is_available(),  # Default
                "remove_unused_columns": True,
                "dataloader_pin_memory": False,
                "gradient_accumulation_steps": accumulate_steps,
                "gradient_checkpointing": True,
                "logging_strategy": "steps",
                "logging_steps": 1,  # logging at every step
                # NOTE: This is explicitly set to false since it will
                # negatively impact the performance
                "full_determinism": False,
                # Required for iterable dataset
                "max_steps": cls.infer_max_steps(
                    num_epochs, batch_size, training_dataset
                ),
                # Some interesting parameters:
                "auto_find_batch_size": True,
                # NOTE: following can override above arguments in order
                **filtered_training_arguments,
                **processing_configuration,
                **dtype_based_params,
            }

            if num_epochs < 1:
                log.warning(
                    "<NLP64076114W>",
                    f"Number of epochs configured is {num_epochs} which is less than minimum 1. \
                        No training will be performed",
                )

                return TextGeneration(
                    model_name=base_model._model_name,
                    model=base_model,
                )

            launch_config = get_torch_elastic_launch_config(
                get_config().master_addr,
                get_config().master_port,
            )

            if torch.cuda.is_available():
                # NOTE: torch distributed can hang if run on CPUs,
                # to avoid that, specially for unit tests, we are only
                # running below when GPUs are available
                training_loss_history = torch.distributed.launcher.api.elastic_launch(
                    launch_config, cls._launch_training
                )(base_model, training_dataset, training_args, checkpoint_dir)

                # NOTE: We are currently only storing the loss information from
                # rank 0, i.e main process. training_loss_history is dictionary containing
                # rank of the process as key
                training_loss_history = training_loss_history[0]
            else:
                training_loss_history = cls._launch_training(
                    base_model, training_dataset, training_args, checkpoint_dir
                )

            # In case this program is started via torchrun, below might not work as is
            # because this case of multiple devices, this whole program gets run
            # in parallel, so the model might still be in "write" mode on 1 process
            # while we try to read it in below process.
            #
            # Note that we forward the base model tokenizer instance to bootstrap to avoid
            # dealing with temporarily exporting and reloading the tokenizer off of the trainer.
            model = resource_type.bootstrap(
                model_name=checkpoint_dir,
                tokenizer_name=checkpoint_dir,
                torch_dtype=torch_dtype,
            )

        return cls(
            model_name=base_model._model_name,
            model=model,
            bos_token=model.tokenizer.bos_token or None,
            sep_token=model.tokenizer.sep_token or None,
            eos_token=model.tokenizer.eos_token or None,
            pad_token=model.tokenizer.pad_token or None,
            training_metadata={"loss": training_loss_history},
        )

    @classmethod
    def load(
        cls,
        model_path: str,
        torch_dtype: str = None,
    ) -> "TextGeneration":
        """Function to load text-generation model

        Args:
            model_path: str
                Path to the model to be loaded.
            torch_dtype: str
                Torch data type to be used when loading the model.
        Returns:
            TextGeneration
                Instance of this class built from the on disk model.
        """

        config = ModuleConfig.load(model_path)

        if torch_dtype is not None:
            torch_dtype = str_to_torch_dtype(torch_dtype)
        elif config.trained_torch_dtype:
            torch_dtype = str_to_torch_dtype(config.trained_torch_dtype)

        base_model_path = config.get("artifact_path")
        error.type_check("<NLP35174683E>", str, base_model_path=base_model_path)

        base_model_path = os.path.join(model_path, base_model_path)
        error.dir_check("<NLP01983374E>", base_model_path)
        return cls.bootstrap(base_model_path, torch_dtype)

    def save(self, model_path):
        """Save caikit model

        Args:
            model_path: str
                Folder to save text-generation caikit model
        """
        saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with saver:
            artifacts_dir = "artifacts"
            saver.update_config(
                {
                    "artifact_path": artifacts_dir,
                    "eos_token": self._eos_token,
                    "torch_dtype": str(self.model._torch_dtype),
                }
            )
            if self.model:
                # This will save both tokenizer and base model
                self.model.save(
                    model_path,
                    tokenizer_dirname=artifacts_dir,
                    base_model_dirname=artifacts_dir,
                )

            training_loss_filename = TRAINING_LOSS_LOG_FILENAME

            saver.update_config({"training_logs": training_loss_filename})

            # We are currently only saving logs containing loss in jsonl format
            if "loss" in self.training_metadata:
                loss_log_lines = self.training_metadata.get("loss")
                error.type_check("<NLP60269855E>", list, loss_log_lines=loss_log_lines)
                with open(
                    os.path.join(model_path, training_loss_filename),
                    "w",
                    encoding="utf-8",
                ) as f:
                    for loss_log in loss_log_lines:
                        loss_log = {"name": "loss", "data": loss_log}
                        json.dump(loss_log, f)
                        f.write("\n")

    @TextGenerationTask.taskmethod()
    def run(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        preserve_input_text: bool = True,
        **kwargs,
    ) -> GeneratedTextResult:

        f"""
        Run the full text generation model.
        Args:
            {GENERATE_FUNCTION_ARGS},
            preserve_input_text: bool
                Applicable only to Causal LLMs.
                Whether or not the source string should be contained in the generated output,
                e.g., as a prefix. Default True. (Source string will appear as prefix)
        Returns:
            GeneratedTextResult
                Generated text result produced by the model.
        """

        # TODO: Beam search currently not supported

        return generate_text_func(
            # Pass HF model
            self.model.model,
            self.model.tokenizer,
            self.PRODUCER_ID,
            self._eos_token,
            text,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            preserve_input_text=preserve_input_text,
            task_type=self.model.TASK_TYPE,
            **kwargs,
        )

    @TokenizationTask.taskmethod()
    def run_tokenizer(
        self,
        text: str,
    ) -> TokenizationResults:
        """Run tokenization task against the model

        Args:
           text: str
                Text to tokenize
        Returns:
            TokenizationResults
                The token count
        """
        raise NotImplementedError("Tokenization not implemented for local")

    ################################## Private Functions ######################################

    @staticmethod
    def _preprocess_function(
        base_model: PretrainedModelBase,
        train_stream: DataStream[GenerationTrainRecord],
        tokenizer: AutoTokenizer,
        max_source_length: int,
        max_target_length: int,
        shuffle: bool,
        use_iterable_dataset: bool,
    ):
        """Pre-process each example to get it prepared for training."""
        dataset_type = TransformersIterableDataset if use_iterable_dataset else Dataset
        log.debug("Loading dataset class: [%s]", dataset_type.__name__)
        fn_kwargs = {
            "tokenizer": tokenizer,
            "max_source_length": max_source_length,
            "max_target_length": max_target_length,
        }
        dataset = dataset_type.from_generator(
            get, gen_kwargs={"train_stream": train_stream}
        )
        mapped_dataset = dataset.map(
            base_model.tokenize_function,
            fn_kwargs=fn_kwargs,
            batched=False,
            # Drop the input / output columns; we need to do this for dimensions to play
            # happily when operating on batched inputs for causal language modeling.
            remove_columns=["input", "output"],
        )

        if shuffle:
            log.debug("Shuffling the dataset")
            return mapped_dataset.shuffle(seed=TextGeneration.RANDOM_SEED)

        return mapped_dataset

    @staticmethod
    def _launch_training(
        base_model, training_dataset, training_args, checkpoint_dir
    ) -> None:
        """Utility function to wrap trainer and execute training"""

        trainer = base_model.get_trainer(
            train_dataset=training_dataset, **training_args
        )

        # Start training via Trainer.train function
        trainer.train()

        # save the model temporarily and reload it
        # this is done, since otherwise the model might be distributed in different
        # devices, in which case its better to use trainer's `prediction_step`
        # functions, but then, they don't always give API similar to `generate`
        # and thus cause incompatibilities in `run` function
        trainer.save_state()
        trainer.save_model(checkpoint_dir)

        # save tokenizer explicitly
        base_model.tokenizer.save_pretrained(checkpoint_dir)

        # Below will return log history but launch will automatically attach rank to it.
        # if started in distributed fashion
        return trainer.state.log_history

    @staticmethod
    def infer_max_steps(
        num_epochs: int,
        batch_size: int,
        training_dataset: Union[Dataset, TransformersIterableDataset],
    ):
        # Calculate the number of samples that we have
        if isinstance(training_dataset, Dataset):
            data_len = len(training_dataset)
        else:
            data_len = 0
            for _ in training_dataset:
                data_len += 1
        # Figure out how many batches we'll have per epoch
        num_batches = data_len // batch_size
        # Assume drop_last=False; in general, this doesn't really matter.
        # We mostly do this to avoid strange behavior when the dataset
        # size is smaller than the batch size.
        if num_batches != (data_len * batch_size):
            num_batches += 1
        num_steps = num_batches * num_epochs
        log.debug("Number of inferred steps: [%s]", num_steps)
        return num_steps


def get(train_stream):
    for data in train_stream:
        yield {"input": data.input, "output": data.output}
